# Copyright (c) 2013-2016 Cisco Systems, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from oslo_db import exception as db_exc
from oslo_log import log as logging
from oslo_utils import excutils
from random import shuffle
import sqlalchemy as sa
from sqlalchemy.orm import aliased
import sqlalchemy.orm.exc as sa_exc
from sqlalchemy.sql import func

from networking_cisco import backwards_compatibility as bc
from networking_cisco.ml2_drivers.nexus import (
    constants as const)
from networking_cisco.ml2_drivers.nexus import (
    exceptions as c_exc)
from networking_cisco.ml2_drivers.nexus import (
    nexus_models_v2)

LOG = logging.getLogger(__name__)


def get_nexusport_binding(port_id, vlan_id, switch_ip, instance_id):
    """Lists a nexusport binding."""
    LOG.debug("get_nexusport_binding() called")
    return _lookup_all_nexus_bindings(port_id=port_id,
                                      vlan_id=vlan_id,
                                      switch_ip=switch_ip,
                                      instance_id=instance_id)


def get_nexus_switchport_binding(port_id, switch_ip):
    """Lists all bindings for this switch & port."""
    LOG.debug("get_nexus_switchport_binding() called")
    return _lookup_all_nexus_bindings(port_id=port_id,
                                      switch_ip=switch_ip)


def get_nexusvlan_binding(vlan_id, switch_ip):
    """Lists a vlan and switch binding."""
    LOG.debug("get_nexusvlan_binding() called")
    return _lookup_all_nexus_bindings(vlan_id=vlan_id, switch_ip=switch_ip)


def get_reserved_bindings(vlan_id, instance_id, switch_ip=None,
                          port_id=None):
    """Lists reserved bindings."""
    LOG.debug("get_reserved_bindings() called")
    if port_id:
        return _lookup_all_nexus_bindings(vlan_id=vlan_id,
                                          switch_ip=switch_ip,
                                          instance_id=instance_id,
                                          port_id=port_id)
    elif switch_ip:
        return _lookup_all_nexus_bindings(vlan_id=vlan_id,
                                          switch_ip=switch_ip,
                                          instance_id=instance_id)
    else:
        return _lookup_all_nexus_bindings(vlan_id=vlan_id,
                                          instance_id=instance_id)


def update_reserved_binding(vlan_id, switch_ip, instance_id,
                            port_id, is_switch_binding=True,
                            is_native=False, ch_grp=0):
    """Updates reserved binding.

    This overloads port bindings to support reserved Switch binding
    used to maintain the state of a switch so it can be viewed by
    all other neutron processes. There's also the case of
    a reserved port binding to keep switch information on a given
    interface.

    The values of these arguments is as follows:
    :param vlan_id: 0
    :param switch_ip: ip address of the switch
    :param instance_id: fixed string RESERVED_NEXUS_SWITCH_DEVICE_ID_R1
    :param port_id: switch-state of ACTIVE, RESTORE_S1, RESTORE_S2, INACTIVE
    :               port-expected port_id
    :param ch_grp:      0 if no port-channel else non-zero integer
    """
    if not port_id:
        LOG.warning("update_reserved_binding called with no state")
        return
    LOG.debug("update_reserved_binding called")
    session = bc.get_writer_session()
    if is_switch_binding:
        # For reserved switch binding
        binding = _lookup_one_nexus_binding(session=session,
                                            vlan_id=vlan_id,
                                            switch_ip=switch_ip,
                                            instance_id=instance_id)
        binding.port_id = port_id
    else:
        # For reserved port binding
        binding = _lookup_one_nexus_binding(session=session,
                                            vlan_id=vlan_id,
                                            switch_ip=switch_ip,
                                            instance_id=instance_id,
                                            port_id=port_id)
    binding.is_native = is_native
    binding.channel_group = ch_grp
    session.merge(binding)
    session.flush()
    return binding


def remove_reserved_binding(vlan_id, switch_ip, instance_id,
                            port_id):
    """Removes reserved binding.

    This overloads port bindings to support reserved Switch binding
    used to maintain the state of a switch so it can be viewed by
    all other neutron processes. There's also the case of
    a reserved port binding to keep switch information on a given
    interface.
    The values of these arguments is as follows:
    :param vlan_id: 0
    :param switch_ip: ip address of the switch
    :param instance_id: fixed string RESERVED_NEXUS_SWITCH_DEVICE_ID_R1
    :param port_id: switch-state of ACTIVE, RESTORE_S1, RESTORE_S2, INACTIVE
    :               port-expected port_id
    """
    if not port_id:
        LOG.warning("remove_reserved_binding called with no state")
        return
    LOG.debug("remove_reserved_binding called")
    session = bc.get_writer_session()
    binding = _lookup_one_nexus_binding(session=session,
                                        vlan_id=vlan_id,
                                        switch_ip=switch_ip,
                                        instance_id=instance_id,
                                        port_id=port_id)
    for bind in binding:
        session.delete(bind)
    session.flush()
    return binding


def get_reserved_switch_binding(switch_ip=None):
    """Get a reserved switch binding."""

    return get_reserved_bindings(
               const.NO_VLAN_OR_VNI_ID,
               const.RESERVED_NEXUS_SWITCH_DEVICE_ID_R1,
               switch_ip)


def add_reserved_switch_binding(switch_ip, state):
    """Add a reserved switch binding."""

    # overload port_id to contain switch state
    add_nexusport_binding(
        state,
        const.NO_VLAN_OR_VNI_ID,
        const.NO_VLAN_OR_VNI_ID,
        switch_ip,
        const.RESERVED_NEXUS_SWITCH_DEVICE_ID_R1)


def update_reserved_switch_binding(switch_ip, state):
    """Update a reserved switch binding."""

    # overload port_id to contain switch state
    update_reserved_binding(
        const.NO_VLAN_OR_VNI_ID,
        switch_ip,
        const.RESERVED_NEXUS_SWITCH_DEVICE_ID_R1,
        state)


def is_reserved_binding(binding):
    """Identifies switch & port operational bindings.

    There are two types of reserved bindings.

    1. The Switch binding purpose is to keep track of the switch state for when
       replay is enabled.  Keeping it in the db, allows for all processes to
       determine known state of each switch.
    2. The reserved port binding is used with baremetal transactions which
       don't rely on host to interface mapping in the ini file.  It is learned
       from the transaction and kept in the data base for further reference.
    """

    return (binding.instance_id in
           [const.RESERVED_NEXUS_SWITCH_DEVICE_ID_R1])


def add_provider_network(network_id, vlan_id):
    session = bc.get_writer_session()
    row = nexus_models_v2.NexusProviderNetwork(network_id=network_id,
                                               vlan_id=vlan_id)
    session.add(row)
    session.flush()
    return row


def delete_provider_network(network_id):
    session = bc.get_writer_session()
    row = session.query(nexus_models_v2.NexusProviderNetwork).filter_by(
        network_id=network_id).one_or_none()
    if row:
        session.delete(row)
        session.flush()


def is_provider_network(network_id):
    session = bc.get_reader_session()
    row = session.query(nexus_models_v2.NexusProviderNetwork).filter_by(
        network_id=network_id).one_or_none()
    return True if row else False


def is_provider_vlan(vlan_id):
    session = bc.get_reader_session()
    row = session.query(nexus_models_v2.NexusProviderNetwork).filter_by(
        vlan_id=vlan_id).one_or_none()
    return True if row else False


def get_nexusport_switch_bindings(switch_ip):
    """Lists all Nexus port switch bindings."""
    LOG.debug("get_nexusport_switch_bindings() called")
    return _lookup_all_nexus_bindings(switch_ip=switch_ip)


def add_nexusport_binding(port_id, vlan_id, vni, switch_ip, instance_id,
                          is_native=False, ch_grp=0):
    """Adds a nexusport binding."""
    LOG.debug("add_nexusport_binding() called")
    session = bc.get_writer_session()
    binding = nexus_models_v2.NexusPortBinding(port_id=port_id,
                  vlan_id=vlan_id,
                  vni=vni,
                  switch_ip=switch_ip,
                  instance_id=instance_id,
                  is_native=is_native,
                  channel_group=ch_grp)
    session.add(binding)
    session.flush()
    return binding


def remove_nexusport_binding(port_id, vlan_id, vni, switch_ip, instance_id):
    """Removes a nexusport binding."""
    LOG.debug("remove_nexusport_binding() called")
    session = bc.get_writer_session()
    binding = _lookup_all_nexus_bindings(session=session,
                                         vlan_id=vlan_id,
                                         vni=vni,
                                         switch_ip=switch_ip,
                                         port_id=port_id,
                                         instance_id=instance_id)
    for bind in binding:
        session.delete(bind)
    session.flush()
    return binding


def update_nexusport_binding(port_id, new_vlan_id):
    """Updates nexusport binding."""
    if not new_vlan_id:
        LOG.warning("update_nexusport_binding called with no vlan")
        return
    LOG.debug("update_nexusport_binding called")
    session = bc.get_writer_session()
    binding = _lookup_one_nexus_binding(session=session, port_id=port_id)
    binding.vlan_id = new_vlan_id
    session.merge(binding)
    session.flush()
    return binding


def remove_all_nexusport_bindings():
    """Removes all nexusport bindings."""

    LOG.debug("remove_all_nexusport_bindings() called")
    session = bc.get_writer_session()
    session.query(nexus_models_v2.NexusPortBinding).delete()
    session.flush()


def get_nexusvm_bindings(vlan_id, instance_id):
    """Lists nexusvm bindings."""
    LOG.debug("get_nexusvm_bindings() called")
    return _lookup_all_nexus_bindings(instance_id=instance_id,
                                      vlan_id=vlan_id)


def get_port_vlan_switch_binding(port_id, vlan_id, switch_ip):
    """Lists nexusvm bindings."""
    LOG.debug("get_port_vlan_switch_binding() called")
    return _lookup_all_nexus_bindings(port_id=port_id,
                                      switch_ip=switch_ip,
                                      vlan_id=vlan_id)


def get_port_switch_bindings(port_id, switch_ip):
    """List all vm/vlan bindings on a Nexus switch port."""
    LOG.debug("get_port_switch_bindings() called, "
              "port:'%(port_id)s', switch:'%(switch_ip)s'",
              {'port_id': port_id, 'switch_ip': switch_ip})
    try:
        return _lookup_all_nexus_bindings(port_id=port_id,
                                          switch_ip=switch_ip)
    except c_exc.NexusPortBindingNotFound:
        pass


def get_nexussvi_bindings():
    """Lists nexus svi bindings."""
    LOG.debug("get_nexussvi_bindings() called")
    return _lookup_all_nexus_bindings(port_id='router')


def _lookup_nexus_bindings(query_type, session=None, **bfilter):
    """Look up 'query_type' Nexus bindings matching the filter.

    :param query_type: 'all', 'one' or 'first'
    :param session: db session
    :param bfilter: filter for bindings query
    :returns: bindings if query gave a result, else
             raise NexusPortBindingNotFound.
    """
    if session is None:
        session = bc.get_reader_session()
    query_method = getattr(session.query(
        nexus_models_v2.NexusPortBinding).filter_by(**bfilter), query_type)
    try:
        bindings = query_method()
        if bindings:
            return bindings
    except sa_exc.NoResultFound:
        pass
    raise c_exc.NexusPortBindingNotFound(**bfilter)


def _lookup_all_nexus_bindings(session=None, **bfilter):
    return _lookup_nexus_bindings('all', session, **bfilter)


def _lookup_one_nexus_binding(session=None, **bfilter):
    return _lookup_nexus_bindings('one', session, **bfilter)


def _lookup_first_nexus_binding(session=None, **bfilter):
    return _lookup_nexus_bindings('first', session, **bfilter)


def add_nexusnve_binding(vni, switch_ip, device_id, mcast_group):
    """Adds a nexus nve binding."""
    LOG.debug("add_nexusnve_binding() called")
    session = bc.get_writer_session()
    binding = nexus_models_v2.NexusNVEBinding(vni=vni,
                                              switch_ip=switch_ip,
                                              device_id=device_id,
                                              mcast_group=mcast_group)
    session.add(binding)
    session.flush()
    return binding


def remove_nexusnve_binding(vni, switch_ip, device_id):
    """Remove the nexus nve binding."""
    LOG.debug("remove_nexusnve_binding() called")
    session = bc.get_writer_session()
    binding = (session.query(nexus_models_v2.NexusNVEBinding).
               filter_by(vni=vni, switch_ip=switch_ip,
                         device_id=device_id).one())
    if binding:
        session.delete(binding)
        session.flush()
        return binding


def remove_all_nexusnve_bindings():
    """Removes all nexusnve bindings."""

    LOG.debug("remove_all_nexusport_bindings() called")
    session = bc.get_writer_session()
    session.query(nexus_models_v2.NexusNVEBinding).delete()
    session.flush()


def get_nve_vni_switch_bindings(vni, switch_ip):
    """Return the nexus nve binding(s) per switch."""
    LOG.debug("get_nve_vni_switch_bindings() called")
    session = bc.get_reader_session()
    try:
        return (session.query(nexus_models_v2.NexusNVEBinding).
                filter_by(vni=vni, switch_ip=switch_ip).all())
    except sa_exc.NoResultFound:
        return None


def get_nve_vni_member_bindings(vni, switch_ip, device_id):
    """Return the nexus nve binding per switch and device_id."""
    LOG.debug("get_nve_vni_member_bindings() called")
    session = bc.get_reader_session()
    try:
        return (session.query(nexus_models_v2.NexusNVEBinding).
                filter_by(vni=vni, switch_ip=switch_ip,
                          device_id=device_id).all())
    except sa_exc.NoResultFound:
        return None


def get_nve_switch_bindings(switch_ip):
    """Return all the nexus nve bindings for one switch."""
    LOG.debug("get_nve_switch_bindings() called")
    session = bc.get_reader_session()
    try:
        return (session.query(nexus_models_v2.NexusNVEBinding).
                filter_by(switch_ip=switch_ip).all())
    except sa_exc.NoResultFound:
        return None


def get_nve_vni_deviceid_bindings(vni, device_id):
    """Return all the nexus nve bindings for one vni/one device_id."""
    LOG.debug("get_nve_vni_deviceid_bindings() called")
    session = bc.get_reader_session()
    try:
        return (session.query(nexus_models_v2.NexusNVEBinding).
                filter_by(vni=vni, device_id=device_id).all())
    except sa_exc.NoResultFound:
        return None


def _lookup_host_mappings(query_type, session=None, **bfilter):
    """Look up 'query_type' Nexus mappings matching the filter.

    :param query_type: 'all', 'one' or 'first'
    :param session: db session
    :param bfilter: filter for mappings query
    :returns: mappings if query gave a result, else
             raise NexusHostMappingNotFound.
    """
    if session is None:
        session = bc.get_reader_session()
    query_method = getattr(session.query(
        nexus_models_v2.NexusHostMapping).filter_by(**bfilter), query_type)
    try:
        mappings = query_method()
        if mappings:
            return mappings
    except sa_exc.NoResultFound:
        pass
    raise c_exc.NexusHostMappingNotFound(**bfilter)


def _lookup_all_host_mappings(session=None, **bfilter):
    return _lookup_host_mappings('all', session, **bfilter)


def _lookup_one_host_mapping(session=None, **bfilter):
    return _lookup_host_mappings('one', session, **bfilter)


def get_all_host_mappings():
    return(_lookup_all_host_mappings())


def get_host_mappings(host_id):
    return(_lookup_all_host_mappings(host_id=host_id))


def get_switch_host_mappings(switch_ip):
    return(_lookup_all_host_mappings(switch_ip=switch_ip))


def get_switch_and_host_mappings(host_id, switch_ip):
    return(_lookup_all_host_mappings(
        host_id=host_id, switch_ip=switch_ip))


def get_switch_if_host_mappings(switch_ip, if_id):
    return(_lookup_all_host_mappings(switch_ip=switch_ip,
                                     if_id=if_id))


def add_host_mapping(host_id, nexus_ip, interface, ch_grp, is_static):
    """Add Host to interface mapping entry into mapping data base.

    :param host_id: is the name of the host to add
    :param interface: is the interface for this host
    :param nexus_ip: is the ip addr of the nexus switch for this interface
    :param ch_grp: is the port channel this interface belos
    :param is_static: whether this is from conf file or learned from baremetal.
    """

    LOG.debug("add_nexusport_binding() called")
    session = bc.get_writer_session()
    mapping = nexus_models_v2.NexusHostMapping(host_id=host_id,
                  if_id=interface,
                  switch_ip=nexus_ip,
                  ch_grp=ch_grp,
                  is_static=is_static)
    try:
        session.add(mapping)
        session.flush()
    except db_exc.DBDuplicateEntry:
        with excutils.save_and_reraise_exception() as ctxt:
            if is_static:
                ctxt.reraise = False
                LOG.debug("Duplicate static entry encountered "
                          "host=%(host)s, if=%(if)s, ip=%(ip)s",
                          {'host': host_id, 'if': interface,
                           'ip': nexus_ip})

    return mapping


def update_host_mapping(host_id, interface, nexus_ip, new_ch_grp):
    """Change channel_group in host/interface mapping data base."""

    LOG.debug("update_host_mapping called")
    session = bc.get_writer_session()
    mapping = _lookup_one_host_mapping(
                  session=session,
                  host_id=host_id,
                  if_id=interface,
                  switch_ip=nexus_ip)
    mapping.ch_grp = new_ch_grp
    session.merge(mapping)
    session.flush()
    return mapping


def remove_host_mapping(interface, nexus_ip):
    """Remove host to interface mapping entry from mapping data base."""

    LOG.debug("remove_host_mapping() called")
    session = bc.get_writer_session()
    try:
        mapping = _lookup_one_host_mapping(
                      session=session,
                      if_id=interface,
                      switch_ip=nexus_ip)
        session.delete(mapping)
        session.flush()
    except c_exc.NexusHostMappingNotFound:
        pass


def remove_all_static_host_mappings():
    """Remove all entries defined in config file from mapping data base."""

    LOG.debug("remove_host_mapping() called")
    session = bc.get_writer_session()
    try:
        mapping = _lookup_all_host_mappings(
                      session=session,
                      is_static=True)
        for host in mapping:
            session.delete(host)
        session.flush()
    except c_exc.NexusHostMappingNotFound:
        pass


def _lookup_vpc_allocs(query_type, session=None, order=None, **bfilter):
    """Look up 'query_type' Nexus VPC Allocs matching the filter.

    :param query_type: 'all', 'one' or 'first'
    :param session: db session
    :param order: select what field to order data
    :param bfilter: filter for mappings query
    :returns: VPCs if query gave a result, else
             raise NexusVPCAllocNotFound.
    """

    if session is None:
        session = bc.get_reader_session()

    if order:
        query_method = getattr(session.query(
            nexus_models_v2.NexusVPCAlloc).filter_by(**bfilter).order_by(
                order),
            query_type)
    else:
        query_method = getattr(session.query(
            nexus_models_v2.NexusVPCAlloc).filter_by(**bfilter), query_type)

    try:
        vpcs = query_method()
        if vpcs:
            return vpcs
    except sa_exc.NoResultFound:
        pass

    raise c_exc.NexusVPCAllocNotFound(**bfilter)


def _lookup_vpc_count_min_max(session=None, **bfilter):
    """Look up count/min/max Nexus VPC Allocs for given switch.

    :param session: db session
    :param bfilter: filter for mappings query
    :returns: number of VPCs and min value if query gave a result,
             else raise NexusVPCAllocNotFound.
    """

    if session is None:
        session = bc.get_reader_session()

    try:
        res = session.query(
            func.count(nexus_models_v2.NexusVPCAlloc.vpc_id),
            func.min(nexus_models_v2.NexusVPCAlloc.vpc_id),
            func.max(nexus_models_v2.NexusVPCAlloc.vpc_id),
        ).filter(nexus_models_v2.NexusVPCAlloc.switch_ip ==
                 bfilter['switch_ip']).one()

        count = res[0]
        sw_min = res[1]
        sw_max = res[2]

        return count, sw_min, sw_max

    except sa_exc.NoResultFound:
        pass

    raise c_exc.NexusVPCAllocNotFound(**bfilter)


def _lookup_all_vpc_allocs(session=None, order=None, **bfilter):
    return _lookup_vpc_allocs('all', session, order, **bfilter)


def _lookup_one_vpc_allocs(session=None, **bfilter):
    return _lookup_vpc_allocs('one', session, **bfilter)


def _get_free_vpcids_on_switches(switch_ip_list):
    '''Get intersect list of free vpcids in list of switches.'''

    session = bc.get_reader_session()

    prev_view = aliased(nexus_models_v2.NexusVPCAlloc)
    query = session.query(prev_view.vpc_id)
    prev_swip = switch_ip_list[0]

    for ip in switch_ip_list[1:]:
        cur_view = aliased(nexus_models_v2.NexusVPCAlloc)
        cur_swip = ip
        query = query.join(cur_view, sa.and_(
            prev_view.switch_ip == prev_swip, prev_view.active == False,  # noqa
            cur_view.switch_ip == cur_swip, cur_view.active == False,     # noqa
            prev_view.vpc_id == cur_view.vpc_id))
        prev_view = cur_view
        prev_swip = cur_swip

    unique_vpcids = query.all()
    shuffle(unique_vpcids)
    return unique_vpcids


def get_all_switch_vpc_allocs(switch_ip):
    try:
        vpc_list = _lookup_all_vpc_allocs(
            order=nexus_models_v2.NexusVPCAlloc.vpc_id,
            switch_ip=switch_ip)
    except c_exc.NexusVPCAllocNotFound:
        vpc_list = []
    return vpc_list


def get_switch_vpc_count_min_max(switch_ip):
    return(_lookup_vpc_count_min_max(switch_ip=switch_ip))


def get_active_switch_vpc_allocs(switch_ip):
    return(_lookup_all_vpc_allocs(switch_ip=switch_ip, active=True))


def get_free_switch_vpc_allocs(switch_ip):
    return(_lookup_all_vpc_allocs(switch_ip=switch_ip, active=False))


def get_switch_vpc_alloc(switch_ip, vpc_id):
    return(_lookup_one_vpc_allocs(switch_ip=switch_ip, vpc_id=vpc_id))


def init_vpc_entries(nexus_ip, vpc_list):
    """Initialize switch/vpc entries in vpc alloc data base.

    param: nexus_ip  ip addr of the nexus switch for this interface
    param: vpc_list  list of vpc integers to create
    """

    LOG.debug("init_vpc_entries() called")

    if not vpc_list:
        return
    session = bc.get_writer_session()

    for vpc in vpc_list:
        vpc_alloc = nexus_models_v2.NexusVPCAlloc(
            switch_ip=nexus_ip,
            vpc_id=vpc,
            learned=False,
            active=False)
        session.add(vpc_alloc)
    session.flush()


def update_vpc_entry(nexus_ips, vpc_id, learned, active):
    """Change active state in vpc_allocate data base."""

    LOG.debug("update_vpc_entry called")

    session = bc.get_writer_session()

    with session.begin():
        for n_ip in nexus_ips:
            flipit = not active
            x = session.execute(
                sa.update(nexus_models_v2.NexusVPCAlloc).values({
                    'learned': learned,
                    'active': active}).where(sa.and_(
                        nexus_models_v2.NexusVPCAlloc.switch_ip == n_ip,
                        nexus_models_v2.NexusVPCAlloc.vpc_id == vpc_id,
                        nexus_models_v2.NexusVPCAlloc.active == flipit
                    )))
            if x.rowcount != 1:
                raise c_exc.NexusVPCAllocNotFound(
                    switch_ip=n_ip, vpc_id=vpc_id, active=active)


def alloc_vpcid(nexus_ips):
    """Allocate a vpc id for the given list of switch_ips."""

    LOG.debug("alloc_vpc() called")

    vpc_id = 0
    intersect = _get_free_vpcids_on_switches(nexus_ips)
    for intersect_tuple in intersect:
        try:
            update_vpc_entry(nexus_ips, intersect_tuple.vpc_id,
                             False, True)
            vpc_id = intersect_tuple.vpc_id
            break
        except Exception:
            # Another controller may have beaten us to this vpcid
            pass

    return vpc_id


def free_vpcid_for_switch_list(vpc_id, nexus_ips):
    """Free a vpc id for the given list of switch_ips."""

    LOG.debug("free_vpcid_for_switch_list() called")
    if vpc_id != 0:
        update_vpc_entry(nexus_ips, vpc_id, False, False)


def free_vpcid_for_switch(vpc_id, nexus_ip):
    """Free a vpc id for the given switch_ip."""

    LOG.debug("free_vpcid_for_switch() called")
    if vpc_id != 0:
        update_vpc_entry([nexus_ip], vpc_id, False, False)


def delete_vpcid_for_switch(vpc_id, switch_ip):
    """Removes unused vpcid for a switch.

    :param vpc_id: vpc id to remove
    :param switch_ip: ip address of the switch
    """

    LOG.debug("delete_vpcid_for_switch called")
    session = bc.get_writer_session()

    vpc = _lookup_one_vpc_allocs(vpc_id=vpc_id,
                                 switch_ip=switch_ip,
                                 active=False)
    session.delete(vpc)
    session.flush()



# Copyright 2014 Cisco Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging
from oslo_utils import excutils

from networking_cisco import backwards_compatibility as bc
from networking_cisco.backwards_compatibility import exnet_const as ext_net
from networking_cisco.plugins.cisco.common import cisco_constants
from networking_cisco.plugins.cisco.device_manager import config
import networking_cisco.plugins.cisco.device_manager.plugging_drivers as plug

LOG = logging.getLogger(__name__)

DEVICE_OWNER_ROUTER_GW = bc.constants.DEVICE_OWNER_ROUTER_GW
GLOBAL_ROUTER_ROLES = [cisco_constants.ROUTER_ROLE_GLOBAL,
                       cisco_constants.ROUTER_ROLE_LOGICAL_GLOBAL]


class HwVLANTrunkingPlugDriver(plug.PluginSidePluggingDriver):
    """Driver class for Cisco hardware-based devices.

    The driver works with VLAN segmented Neutron networks.
    """
    # once initialized _device_network_interface_map is dictionary
    _device_network_interface_map = None

    @property
    def _core_plugin(self):
        try:
            return self._plugin
        except AttributeError:
            self._plugin = bc.get_plugin()
            return self._plugin

    def create_hosting_device_resources(self, context, complementary_id,
                                        tenant_id, mgmt_context, max_hosted):
        return {'mgmt_port': None}

    def get_hosting_device_resources(self, context, id, complementary_id,
                                     tenant_id, mgmt_nw_id):
        return {'mgmt_port': None,
                'ports': [], 'networks': [], 'subnets': []}

    def delete_hosting_device_resources(self, context, tenant_id, mgmt_port,
                                        **kwargs):
        pass

    def setup_logical_port_connectivity(self, context, port_db,
                                        hosting_device_id):
        pass

    def teardown_logical_port_connectivity(self, context, port_db,
                                           hosting_device_id):
        pass

    def extend_hosting_port_info(self, context, port_db, hosting_device,
                                 hosting_info):
        hosting_info['segmentation_id'] = port_db.hosting_info.segmentation_id
        if bc.NEUTRON_VERSION < bc.NEUTRON_PIKE_VERSION:
            is_external = (port_db.networks.external is not None)
        else:
            nw = self._core_plugin.get_network(context, port_db.network_id)
            is_external = nw[ext_net.EXTERNAL]
        hosting_info['physical_interface'] = self._get_interface_info(
            hosting_device['id'], port_db.network_id, is_external)

    def allocate_hosting_port(self, context, router_id, port_db, network_type,
                              hosting_device_id):
        # For VLAN core plugin provides VLAN tag
        tags = self._core_plugin.get_networks(
            context, {'id': [port_db['network_id']]},
            [bc.provider_net.SEGMENTATION_ID])
        allocated_vlan = (None if tags == []
                          else tags[0].get(bc.provider_net.SEGMENTATION_ID))
        if allocated_vlan is None:
            # Database must have been messed up if this happens ...
            LOG.debug('hw_vlan_trunking_driver: Could not allocate VLAN')
            return
        return {'allocated_port_id': port_db.id,
                'allocated_vlan': allocated_vlan}

    @classmethod
    def _get_interface_info(cls, device_id, network_id, external=False):
        if cls._device_network_interface_map is None:
            cls._get_network_interface_map_from_config()
        try:
            dev_info = cls._device_network_interface_map[device_id]
            if external:
                return dev_info['external'].get(network_id,
                                                dev_info['external']['*'])
            else:
                return dev_info['internal'].get(network_id,
                                                dev_info['internal']['*'])
        except (TypeError, KeyError):
            LOG.error('Failed to lookup interface on device %(dev)s'
                      'for network %(net)s', {'dev': device_id,
                                              'net': network_id})
            return

    @classmethod
    def _get_network_interface_map_from_config(cls):
        dni_dict = config.get_specific_config(
            'HwVLANTrunkingPlugDriver'.lower())
        temp = {}
        for hd_uuid, kv_dict in dni_dict.items():
            # ensure hd_uuid is properly formatted
            hd_uuid = config.uuidify(hd_uuid)
            if hd_uuid not in temp:
                temp[hd_uuid] = {'internal': {}, 'external': {}}
            for k, v in kv_dict.items():
                try:
                    entry = k[:k.index('_')]
                    net_spec, interface = v.split(':')
                    for net_id in net_spec.split(','):
                        temp[hd_uuid][entry][net_id] = interface
                except (ValueError, KeyError):
                    with excutils.save_and_reraise_exception() as ctx:
                        ctx.reraise = False
                        LOG.error('Invalid network to interface mapping '
                                  '%(key)s, %(value)s in configuration '
                                  'file for device = %(dev)s',
                                  {'key': k, 'value': v, 'dev': hd_uuid})
        cls._device_network_interface_map = temp



# Copyright 2014 Cisco Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os

import contextlib
import mock

from neutron.api import extensions as api_ext
from neutron.common import config
from neutron.tests.unit.db import test_db_base_plugin_v2
from oslo_config import cfg
from oslo_utils import importutils
import six
import webob.exc

import networking_cisco
from networking_cisco import backwards_compatibility as bc
from networking_cisco.plugins.cisco.common import (cisco_constants as
                                                   c_constants)
from networking_cisco.plugins.cisco.db.device_manager import (
    hosting_device_manager_db as hdm_db)
from networking_cisco.plugins.cisco.device_manager.rpc import (
    devmgr_rpc_cfgagent_api)
from networking_cisco.plugins.cisco.device_manager import service_vm_lib
from networking_cisco.plugins.cisco.extensions import ciscohostingdevicemanager
from networking_cisco.tests.unit.cisco.device_manager import (
    device_manager_test_support)


policy_path = (os.path.abspath(networking_cisco.__path__[0]) +
               '/../etc/policy.json')

DB_DM_PLUGIN_KLASS = (
    'networking_cisco.plugins.cisco.db.device_manager.'
    'hosting_device_manager_db.HostingDeviceManagerMixin')

NN_CATEGORY = ciscohostingdevicemanager.NETWORK_NODE_CATEGORY
NN_TEMPLATE_NAME = c_constants.NETWORK_NODE_TEMPLATE
NS_ROUTERTYPE_NAME = c_constants.NAMESPACE_ROUTER_TYPE
VM_CATEGORY = ciscohostingdevicemanager.VM_CATEGORY
VM_TEMPLATE_NAME = "VM_template"
VM_BOOTING_TIME = 420
VM_SLOT_CAPACITY = 3
VM_DESIRED_SLOTS_FREE = 3
VM_ROUTERTYPE_NAME = "VM_router"
HW_CATEGORY = ciscohostingdevicemanager.HARDWARE_CATEGORY
HW_TEMPLATE_NAME = "HW_template"
HW_ROUTERTYPE_NAME = c_constants.ASR1K_ROUTER_TYPE

L3_ROUTER_NAT = bc.constants.L3

DEFAULT_SERVICE_TYPES = "router"
NETWORK_NODE_SERVICE_TYPES = "router:fwaas:vpn"

NOOP_DEVICE_DRIVER = ('networking_cisco.plugins.cisco.device_manager.'
                      'hosting_device_drivers.noop_hd_driver.'
                      'NoopHostingDeviceDriver')
NOOP_PLUGGING_DRIVER = ('networking_cisco.plugins.cisco.device_manager.'
                        'plugging_drivers.noop_plugging_driver.'
                        'NoopPluggingDriver')

TEST_DEVICE_DRIVER = NOOP_DEVICE_DRIVER
TEST_PLUGGING_DRIVER = ('networking_cisco.tests.unit.cisco.device_manager.'
                        'plugging_test_driver.TestPluggingDriver')

DESCRIPTION = "default description"
SHARED = True
ACTION = "allow"
ENABLED = True
ADMIN_STATE_UP = True

UNBOUND = None
REQUESTER = True
OTHER = False

DEFAULT_CREDENTIALS_ID = device_manager_test_support._uuid()


class DeviceManagerTestCaseMixin(object):

    def _create_hosting_device(self, fmt, template_id, management_port_id,
                               admin_state_up, expected_res_status=None,
                               **kwargs):
        data = {'hosting_device': self._get_test_hosting_device_attr(
            template_id=template_id, management_port_id=management_port_id,
            admin_state_up=admin_state_up, **kwargs)}
        hd_req = self.new_create_request('hosting_devices', data, fmt)
        if kwargs.get('set_context') and 'tenant_id' in kwargs:
            # create a specific auth context for this request
            hd_req.environ['neutron.context'] = bc.context.Context(
                '', kwargs['tenant_id'])
        hd_res = hd_req.get_response(self.ext_api)
        if expected_res_status:
            self.assertEqual(expected_res_status, hd_res.status_int)
        return hd_res

    @contextlib.contextmanager
    def hosting_device(self, template_id, management_port_id=None, fmt=None,
                       admin_state_up=True, no_delete=False,
                       set_port_device_id=True, **kwargs):
        if not fmt:
            fmt = self.fmt
        res = self._create_hosting_device(fmt, template_id, management_port_id,
                                          admin_state_up, **kwargs)
        if res.status_int >= 400:
            raise webob.exc.HTTPClientError(code=res.status_int)
        hosting_device = self.deserialize(fmt or self.fmt, res)
        if set_port_device_id is True and management_port_id is not None:
            data = {'port': {
                'device_id': hosting_device['hosting_device']['id'],
                'device_owner': 'Nova'}}
            req = self.new_update_request('ports', data, management_port_id)
            res = self.deserialize(self.fmt, req.get_response(self.api))
        yield hosting_device
        if not no_delete:
            self._delete('hosting_devices',
                         hosting_device['hosting_device']['id'])

    def _create_hosting_device_template(self, fmt, name, enabled,
                                        host_category,
                                        expected_res_status=None, **kwargs):
        data = {'hosting_device_template':
                self._get_test_hosting_device_template_attr(
                    name=name, enabled=enabled, host_category=host_category,
                    **kwargs)}
        hdt_req = self.new_create_request('hosting_device_templates', data,
                                          fmt)
        if kwargs.get('set_context') and 'tenant_id' in kwargs:
            # create a specific auth context for this request
            hdt_req.environ['neutron.context'] = bc.context.Context(
                '', kwargs['tenant_id'])
        hdt_res = hdt_req.get_response(self.ext_api)
        if expected_res_status:
            self.assertEqual(expected_res_status, hdt_res.status_int)
        return hdt_res

    @contextlib.contextmanager
    def hosting_device_template(self, fmt=None, name='device_template_1',
                                enabled=True, host_category=VM_CATEGORY,
                                no_delete=False, **kwargs):
        if not fmt:
            fmt = self.fmt
        res = self._create_hosting_device_template(fmt, name, enabled,
                                                   host_category, **kwargs)
        if res.status_int >= 400:
            raise webob.exc.HTTPClientError(code=res.status_int)
        hd_template = self.deserialize(fmt or self.fmt, res)
        yield hd_template
        if not no_delete:
            self._delete('hosting_device_templates',
                         hd_template['hosting_device_template']['id'])

    def _get_test_hosting_device_attr(self, template_id, management_port_id,
                                      admin_state_up=True, **kwargs):
        data = {
            'tenant_id': kwargs.get('tenant_id', self._tenant_id),
            'template_id': template_id,
            'credentials_id': kwargs.get('credentials_id'),
            'device_id': kwargs.get('device_id', 'mfc_device_id'),
            'admin_state_up': admin_state_up,
            'management_ip_address': kwargs.get('management_ip_address',
                                                '10.0.100.10'),
            'management_port_id': management_port_id,
            'protocol_port': kwargs.get('protocol_port', 22),
            'cfg_agent_id': kwargs.get('cfg_agent_id'),
            'tenant_bound': kwargs.get('tenant_bound'),
            'auto_delete': kwargs.get('auto_delete', False)}
        return data

    def _get_test_hosting_device_template_attr(self, name='device_template_1',
                                               enabled=True,
                                               host_category=VM_CATEGORY,
                                               **kwargs):
        data = {
            'tenant_id': kwargs.get('tenant_id', self._tenant_id),
            'name': name,
            'enabled': enabled,
            'host_category': host_category,
            'service_types': kwargs.get('service_types',
                                        DEFAULT_SERVICE_TYPES),
            'image': kwargs.get('image'),
            'flavor': kwargs.get('flavor'),
            'default_credentials_id': kwargs.get('default_credentials_id',
                                                 DEFAULT_CREDENTIALS_ID),
            'configuration_mechanism': kwargs.get('configuration_mechanism'),
            'protocol_port': kwargs.get('protocol_port', 22),
            'booting_time': kwargs.get('booting_time', 0),
            'slot_capacity': kwargs.get('slot_capacity', 0),
            'desired_slots_free': kwargs.get('desired_slots_free', 0),
            'tenant_bound': kwargs.get('tenant_bound', []),
            'device_driver': kwargs.get('device_driver', NOOP_DEVICE_DRIVER),
            'plugging_driver': kwargs.get('plugging_driver',
                                          NOOP_PLUGGING_DRIVER)}
        return data

    def _test_list_resources(self, resource, items,
                             neutron_context=None,
                             query_params=None):
        if resource.endswith('y'):
            resource_plural = resource.replace('y', 'ies')
        else:
            resource_plural = resource + 's'

        res = self._list(resource_plural,
                         neutron_context=neutron_context,
                         query_params=query_params)
        resource = resource.replace('-', '_')
        self.assertEqual(sorted([i[resource]['id'] for i in items]),
                         sorted([i['id'] for i in res[resource_plural]]))

    def _replace_hosting_device_status(self, attrs, old_status, new_status):
        if attrs['status'] is old_status:
            attrs['status'] = new_status
        return attrs

    def _test_create_hosting_device_templates(self):
        # template for network nodes.
        nnt = self._create_hosting_device_template(self.fmt, NN_TEMPLATE_NAME,
                                                   True, NN_CATEGORY)
        nw_node_template = self.deserialize(self.fmt, nnt)
        vmt = self._create_hosting_device_template(
            self.fmt, VM_TEMPLATE_NAME, True, VM_CATEGORY,
            booting_time=VM_BOOTING_TIME,
            slot_capacity=VM_SLOT_CAPACITY,
            desired_slots_free=VM_DESIRED_SLOTS_FREE,
            device_driver=TEST_DEVICE_DRIVER,
            plugging_driver=TEST_PLUGGING_DRIVER)
        vm_template = self.deserialize(self.fmt, vmt)
        hwt = self._create_hosting_device_template(
            self.fmt, HW_TEMPLATE_NAME, True, HW_CATEGORY)
        hw_template = self.deserialize(self.fmt, hwt)
        return {'network_node': {'template': nw_node_template,
                                 'router_type': NS_ROUTERTYPE_NAME},
                'vm': {'template': vm_template,
                       'router_type': VM_ROUTERTYPE_NAME},
                'hw': {'template': hw_template,
                       'router_type': HW_ROUTERTYPE_NAME}}

    def _test_remove_hosting_device_templates(self):
        for hdt in self._list('hosting_device_templates')[
                'hosting_device_templates']:
            self._delete('hosting_device_templates', hdt['id'])


class TestDeviceManagerDBPlugin(
    test_db_base_plugin_v2.NeutronDbPluginV2TestCase,
    DeviceManagerTestCaseMixin,
        device_manager_test_support.DeviceManagerTestSupportMixin):

    hdm_db.HostingDeviceManagerMixin.path_prefix = "/dev_mgr"
    resource_prefix_map = dict(
        (k, "/dev_mgr")
        for k in ciscohostingdevicemanager.RESOURCE_ATTRIBUTE_MAP.keys())

    def setUp(self, core_plugin=None, dm_plugin=None, ext_mgr=None):
        if dm_plugin is None:
            dm_plugin = DB_DM_PLUGIN_KLASS
        service_plugins = {'dm_plugin_name': dm_plugin}
        cfg.CONF.set_override('api_extensions_path',
                              device_manager_test_support.extensions_path)
        # for these tests we need to enable overlapping ips
        cfg.CONF.set_default('allow_overlapping_ips', True)
        hdm_db.HostingDeviceManagerMixin.supported_extension_aliases = (
            [ciscohostingdevicemanager.HOSTING_DEVICE_MANAGER_ALIAS])
        super(TestDeviceManagerDBPlugin, self).setUp(
            plugin=core_plugin, service_plugins=service_plugins,
            ext_mgr=ext_mgr)
        # Ensure we use policy definitions from our repo
        cfg.CONF.set_override('policy_file', policy_path, 'oslo_policy')
        if not ext_mgr:
            self.plugin = importutils.import_object(dm_plugin)
            ext_mgr = api_ext.PluginAwareExtensionManager(
                device_manager_test_support.extensions_path,
                {c_constants.DEVICE_MANAGER: self.plugin})
            app = config.load_paste_app('extensions_test_app')
            self.ext_api = api_ext.ExtensionMiddleware(app, ext_mgr=ext_mgr)

        self._mock_l3_admin_tenant()
        self._create_mgmt_nw_for_tests(self.fmt)
        self._devmgr = bc.get_plugin(c_constants.DEVICE_MANAGER)
        # in unit tests we don't use keystone so we mock that session
        self._devmgr._svc_vm_mgr_obj = service_vm_lib.ServiceVMManager(
            True, None, None, None, '', keystone_session=mock.MagicMock())
        self._mock_svc_vm_create_delete(self._devmgr)
        self._other_tenant_id = device_manager_test_support._uuid()
        self._devmgr._core_plugin = bc.get_plugin()

    def tearDown(self):
        self._test_remove_all_hosting_devices()
        self._remove_mgmt_nw_for_tests()
        super(TestDeviceManagerDBPlugin, self).tearDown()

    def test_create_vm_hosting_device(self):
        with self.hosting_device_template() as hdt:
            with self.port(subnet=self._mgmt_subnet) as mgmt_port:
                creds = device_manager_test_support._uuid()
                attrs = self._get_test_hosting_device_attr(
                    template_id=hdt['hosting_device_template']['id'],
                    management_port_id=mgmt_port['port']['id'],
                    auto_delete=True, credentials_id=creds)
                with self.hosting_device(
                        template_id=hdt['hosting_device_template']['id'],
                        management_port_id=mgmt_port['port']['id'],
                        auto_delete=True, credentials_id=creds) as hd:
                    for k, v in six.iteritems(attrs):
                        self.assertEqual(v, hd['hosting_device'][k])

    def test_create_hw_hosting_device(self):
        with self.hosting_device_template(host_category=HW_CATEGORY) as hdt:
            with self.port(subnet=self._mgmt_subnet) as mgmt_port:
                creds = device_manager_test_support._uuid()
                attrs = self._get_test_hosting_device_attr(
                    template_id=hdt['hosting_device_template']['id'],
                    management_port_id=mgmt_port['port']['id'],
                    credentials_id=creds)
                with self.hosting_device(
                        template_id=hdt['hosting_device_template']['id'],
                        management_port_id=mgmt_port['port']['id'],
                        credentials_id=creds) as hd:
                    for k, v in six.iteritems(attrs):
                        self.assertEqual(v, hd['hosting_device'][k])

    def test_show_hosting_device(self):
        device_id = "device_XYZ"
        with self.hosting_device_template() as hdt:
            with self.port(subnet=self._mgmt_subnet) as mgmt_port:
                creds = device_manager_test_support._uuid()
                attrs = self._get_test_hosting_device_attr(
                    device_id=device_id,
                    template_id=hdt['hosting_device_template']['id'],
                    management_port_id=mgmt_port['port']['id'],
                    credentials_id=creds)
                with self.hosting_device(
                        device_id=device_id,
                        template_id=hdt['hosting_device_template']['id'],
                        management_port_id=mgmt_port['port']['id'],
                        credentials_id=creds) as hd:
                    req = self.new_show_request(
                        'hosting_devices', hd['hosting_device']['id'],
                        fmt=self.fmt)
                    res = self.deserialize(self.fmt,
                                           req.get_response(self.ext_api))
                    for k, v in six.iteritems(attrs):
                        self.assertEqual(v, res['hosting_device'][k])

    def test_list_hosting_devices(self):
        with self.hosting_device_template() as hdt:
            hdt_id = hdt['hosting_device_template']['id']
            with self.port(subnet=self._mgmt_subnet) as mgmt_port1,\
                    self.port(subnet=self._mgmt_subnet) as mgmt_port2,\
                    self.port(subnet=self._mgmt_subnet) as mgmt_port3:
                mp1_id = mgmt_port1['port']['id']
                mp2_id = mgmt_port2['port']['id']
                mp3_id = mgmt_port3['port']['id']
                with self.hosting_device(
                        name='hd1', template_id=hdt_id,
                        management_port_id=mp1_id) as hd1,\
                        self.hosting_device(
                            name='hd2', template_id=hdt_id,
                            management_port_id=mp2_id) as hd2,\
                        self.hosting_device(
                            name='hd3', template_id=hdt_id,
                            management_port_id=mp3_id) as hd3:
                    self._test_list_resources(
                        'hosting_device', [hd1, hd2, hd3],
                        query_params='template_id=' + hdt_id)

    def test_update_hosting_device(self):
        new_device_id = "device_XYZ"
        with self.hosting_device_template() as hdt:
            hdt_id = hdt['hosting_device_template']['id']
            with self.port(subnet=self._mgmt_subnet) as mgmt_port:
                mgmt_port_id = mgmt_port['port']['id']
                creds = device_manager_test_support._uuid()
                attrs = self._get_test_hosting_device_attr(
                    device_id=new_device_id,
                    template_id=hdt['hosting_device_template']['id'],
                    management_port_id=mgmt_port['port']['id'],
                    credentials_id=creds)
                with self.hosting_device(
                        template_id=hdt_id,
                        management_port_id=mgmt_port_id,
                        credentials_id=creds) as hd:
                    data = {'hosting_device': {'device_id': new_device_id}}
                    req = self.new_update_request('hosting_devices', data,
                                                  hd['hosting_device']['id'])
                    res = self.deserialize(self.fmt,
                                           req.get_response(self.ext_api))
                    for k, v in six.iteritems(attrs):
                        self.assertEqual(v, res['hosting_device'][k])

    def test_delete_hosting_device_not_in_use_succeeds(self):
        ctx = bc.context.get_admin_context()
        with self.hosting_device_template() as hdt:
            hdt_id = hdt['hosting_device_template']['id']
            with self.port(subnet=self._mgmt_subnet) as mgmt_port:
                mgmt_port_id = mgmt_port['port']['id']
                with self.hosting_device(template_id=hdt_id,
                                         management_port_id=mgmt_port_id,
                                         no_delete=True) as hd:
                    hd_id = hd['hosting_device']['id']
                    req = self.new_delete_request('hosting_devices', hd_id)
                    res = req.get_response(self.ext_api)
                    self.assertEqual(204, res.status_int)
                    self.assertRaises(
                        ciscohostingdevicemanager.HostingDeviceNotFound,
                        self.plugin.get_hosting_device, ctx, hd_id)

    def test_delete_hosting_device_in_use_fails(self):
        ctx = bc.context.get_admin_context()
        with self.hosting_device_template(slot_capacity=1) as hdt:
            hdt_id = hdt['hosting_device_template']['id']
            with self.port(subnet=self._mgmt_subnet) as mgmt_port:
                mgmt_port_id = mgmt_port['port']['id']
                with self.hosting_device(
                        template_id=hdt_id,
                        management_port_id=mgmt_port_id) as hd:
                    with mock.patch.object(
                            hdm_db.HostingDeviceManagerMixin,
                            '_dispatch_pool_maintenance_job'):
                        hd_id = hd['hosting_device']['id']
                        hd_db = self._devmgr._get_hosting_device(ctx, hd_id)
                        resource = self._get_fake_resource()
                        self.assertTrue(
                            self._devmgr.acquire_hosting_device_slots(
                                ctx, hd_db, resource, 'router', L3_ROUTER_NAT,
                                1))
                        self.assertRaises(
                            ciscohostingdevicemanager.HostingDeviceInUse,
                            self._devmgr.delete_hosting_device, ctx, hd_id)
                        req = self.new_show_request('hosting_devices', hd_id,
                                                    fmt=self.fmt)
                        res = req.get_response(self.ext_api)
                        self.assertEqual(200, res.status_int)
                        self._devmgr.release_hosting_device_slots(ctx, hd_db,
                                                                  resource, 1)

    def test_get_hosting_device_configuration(self):
        with self.hosting_device_template() as hdt:
            hdt_id = hdt['hosting_device_template']['id']
            with self.port(subnet=self._mgmt_subnet) as mgmt_port:
                mgmt_port_id = mgmt_port['port']['id']
                with self.hosting_device(
                        template_id=hdt_id,
                        management_port_id=mgmt_port_id) as hd:
                    hd_id = hd['hosting_device']['id']
                    rpc = devmgr_rpc_cfgagent_api.DeviceMgrCfgAgentNotifyAPI(
                        self._devmgr)
                    self._devmgr.agent_notifiers = {
                        c_constants.AGENT_TYPE_CFG: rpc}
                    self._devmgr.get_cfg_agents_for_hosting_devices = None
                    with mock.patch.object(rpc.client, 'prepare',
                                           return_value=rpc.client) as (
                            mock_prepare),\
                        mock.patch.object(rpc.client, 'call') as mock_call,\
                        mock.patch.object(
                            self._devmgr,
                            'get_cfg_agents_for_hosting_devices') as agt_mock:
                        agt_mock.return_value = [mock.MagicMock()]
                        agent_host = 'an_agent_host'
                        agt_mock.return_value[0].host = agent_host
                        fake_running_config = 'a fake running config'
                        mock_call.return_value = fake_running_config
                        ctx = bc.context.Context(
                            user_id=None, tenant_id=None, is_admin=False,
                            overwrite=False)
                        res = self._devmgr.get_hosting_device_config(ctx,
                                                                     hd_id)
                        self.assertEqual(fake_running_config, res)
                        agt_mock.assert_called_once_with(
                            mock.ANY, [hd_id], admin_state_up=True,
                            schedule=True)
                        mock_prepare.assert_called_with(server=agent_host)
                        mock_call.assert_called_with(
                            mock.ANY, 'get_hosting_device_configuration',
                            payload={'hosting_device_id': hd_id})

    def test_get_hosting_device_configuration_no_agent_found(self):
        ctx = bc.context.Context(user_id=None, tenant_id=None, is_admin=False,
                                overwrite=False)
        with self.hosting_device_template() as hdt:
            hdt_id = hdt['hosting_device_template']['id']
            with self.port(subnet=self._mgmt_subnet) as mgmt_port:
                mgmt_port_id = mgmt_port['port']['id']
                with self.hosting_device(
                        template_id=hdt_id,
                        management_port_id=mgmt_port_id) as hd:
                    hd_id = hd['hosting_device']['id']
                    rpc = devmgr_rpc_cfgagent_api.DeviceMgrCfgAgentNotifyAPI(
                        self._devmgr)
                    self._devmgr.agent_notifiers = {
                        c_constants.AGENT_TYPE_CFG: rpc}
                    self._devmgr.get_cfg_agents_for_hosting_devices = None
                    with mock.patch.object(rpc.client, 'prepare',
                                           return_value=rpc.client) as (
                            mock_prepare),\
                        mock.patch.object(rpc.client, 'call') as mock_call,\
                        mock.patch.object(
                            self._devmgr,
                            'get_cfg_agents_for_hosting_devices') as agt_mock:
                        agt_mock.return_value = []
                        res = self._devmgr.get_hosting_device_config(ctx,
                                                                     hd_id)
                        self.assertIsNone(res)
                        agt_mock.assert_called_once_with(
                            mock.ANY, [hd_id], admin_state_up=True,
                            schedule=True)
                        self.assertEqual(0, mock_prepare.call_count)
                        self.assertEqual(0, mock_call.call_count)

    def test_hosting_device_policy(self):
        device_id = "device_XYZ"
        if (bc.NEUTRON_VERSION < bc.NEUTRON_OCATA_VERSION):
            expect_code = webob.exc.HTTPForbidden.code
        else:
            expect_code = webob.exc.HTTPNotFound.code
        with self.hosting_device_template() as hdt:
            hdt_id = hdt['hosting_device_template']['id']
            tenant_id = hdt['hosting_device_template']['tenant_id']
            with self.port(subnet=self._mgmt_subnet) as mgmt_port:
                mgmt_port_id = mgmt_port['port']['id']
                creds = device_manager_test_support._uuid()
                with self.hosting_device(
                        device_id=device_id,
                        template_id=hdt_id,
                        management_port_id=mgmt_port_id,
                        credentials_id=creds) as hd:
                    hd_id = hd['hosting_device']['id']
                    # create fails
                    self._create_hosting_device(
                        self.fmt, hdt_id, mgmt_port_id, True,
                        webob.exc.HTTPForbidden.code,
                        tenant_id=tenant_id, set_context=True)
                    non_admin_ctx = bc.context.Context('', tenant_id)
                    # show fails
                    self._show('hosting_devices', hd_id,
                               webob.exc.HTTPNotFound.code, non_admin_ctx)
                    # update fails
                    self._update('hosting_devices', hd_id,
                                 {'hosting_device': {'name': 'new_name'}},
                                 expect_code, non_admin_ctx)
                    # delete fails
                    self._delete('hosting_devices', hd_id,
                                 webob.exc.HTTPNotFound.code, non_admin_ctx)
                    # get config fails
                    req = self.new_show_request(
                        'hosting_devices', hd_id, self.fmt,
                        'get_hosting_device_config')
                    req.environ['neutron.context'] = non_admin_ctx
                    res = req.get_response(self._api_for_resource(
                        'hosting_devices'))
                    self.assertEqual(webob.exc.HTTPNotFound.code,
                                     res.status_int)

    def test_create_vm_hosting_device_template(self):
        attrs = self._get_test_hosting_device_template_attr()

        with self.hosting_device_template() as hdt:
            for k, v in six.iteritems(attrs):
                self.assertEqual(v, hdt['hosting_device_template'][k])

    def test_create_hw_hosting_device_template(self):
        attrs = self._get_test_hosting_device_template_attr(
            host_category=HW_CATEGORY)

        with self.hosting_device_template(host_category=HW_CATEGORY) as hdt:
            for k, v in six.iteritems(attrs):
                self.assertEqual(v, hdt['hosting_device_template'][k])

    def test_create_nn_hosting_device_template(self):
        attrs = self._get_test_hosting_device_template_attr(
            host_category=NN_CATEGORY)

        with self.hosting_device_template(host_category=NN_CATEGORY) as hdt:
            for k, v in six.iteritems(attrs):
                self.assertEqual(v, hdt['hosting_device_template'][k])

    def test_show_hosting_device_template(self):
        name = "hosting_device_template1"
        attrs = self._get_test_hosting_device_template_attr(name=name)
        with self.hosting_device_template(name=name) as hdt:
            req = self.new_show_request('hosting_device_templates',
                                        hdt['hosting_device_template']['id'],
                                        fmt=self.fmt)
            res = self.deserialize(self.fmt,
                                   req.get_response(self.ext_api))
            for k, v in six.iteritems(attrs):
                self.assertEqual(v, res['hosting_device_template'][k])

    def test_list_hosting_device_templates(self):
        with self.hosting_device_template(name='hdt1',
                                          host_category=VM_CATEGORY,
                                          image='an_image') as hdt1,\
                self.hosting_device_template(name='hdt2',
                                             host_category=HW_CATEGORY,
                                             image='an_image') as hdt2,\
                self.hosting_device_template(name='hdt3',
                                             host_category=NN_CATEGORY,
                                             image='an_image') as hdt3:
            self._test_list_resources(
                'hosting_device_template', [hdt1, hdt2, hdt3],
                query_params='image=an_image')

    def test_update_hosting_device_template(self):
        name = "new_hosting_device_template1"
        attrs = self._get_test_hosting_device_template_attr(name=name)
        with self.hosting_device_template() as hdt:
            data = {'hosting_device_template': {'name': name}}
            req = self.new_update_request('hosting_device_templates', data,
                                          hdt['hosting_device_template']['id'])
            res = self.deserialize(self.fmt,
                                   req.get_response(self.ext_api))
            for k, v in six.iteritems(attrs):
                self.assertEqual(v, res['hosting_device_template'][k])

    def test_delete_hosting_device_template_not_in_use_succeeds(self):
        ctx = bc.context.get_admin_context()
        with self.hosting_device_template(no_delete=True) as hdt:
            hdt_id = hdt['hosting_device_template']['id']
            req = self.new_delete_request('hosting_device_templates', hdt_id)
            res = req.get_response(self.ext_api)
            self.assertEqual(204, res.status_int)
            self.assertRaises(
                ciscohostingdevicemanager.HostingDeviceTemplateNotFound,
                self._devmgr.get_hosting_device_template, ctx, hdt_id)

    def test_delete_hosting_device_template_in_use_fails(self):
        ctx = bc.context.get_admin_context()
        with self.hosting_device_template() as hdt:
            hdt_id = hdt['hosting_device_template']['id']
            with self.port(subnet=self._mgmt_subnet) as mgmt_port:
                mgmt_port_id = mgmt_port['port']['id']
                with self.hosting_device(template_id=hdt_id,
                                         management_port_id=mgmt_port_id):
                    self.assertRaises(
                        ciscohostingdevicemanager.HostingDeviceTemplateInUse,
                        self._devmgr.delete_hosting_device_template, ctx,
                        hdt_id)
                    req = self.new_show_request('hosting_device_templates',
                                                hdt_id, fmt=self.fmt)
                    res = req.get_response(self.ext_api)
                    self.assertEqual(200, res.status_int)

    def test_hosting_device_template_policy(self):
        with self.hosting_device_template() as hdt:
            hdt_id = hdt['hosting_device_template']['id']
            tenant_id = hdt['hosting_device_template']['tenant_id']
            if (bc.NEUTRON_VERSION < bc.NEUTRON_OCATA_VERSION):
                expect_code = webob.exc.HTTPForbidden.code
            else:
                expect_code = webob.exc.HTTPNotFound.code
            # create fails
            self._create_hosting_device_template(
                self.fmt, 'my_template', True, 'Hardware',
                webob.exc.HTTPForbidden.code,
                tenant_id=tenant_id, set_context=True)
            non_admin_ctx = bc.context.Context('', tenant_id)
            # show fail
            self._show('hosting_device_templates', hdt_id,
                       webob.exc.HTTPNotFound.code, non_admin_ctx)
            # update fail
            self._update('hosting_device_templates', hdt_id,
                         {'hosting_device_template': {'enabled': False}},
                         expect_code, non_admin_ctx)
            # delete fail
            self._delete('hosting_device_templates', hdt_id,
                         webob.exc.HTTPNotFound.code, non_admin_ctx)

    # driver request test helper
    def _test_get_driver(self, get_method, id=None, test_for_none=False,
                         is_admin=False):
        with self.hosting_device_template() as hdt:
            context = self._get_test_context(
                tenant_id=hdt['hosting_device_template']['tenant_id'],
                is_admin=is_admin)
            driver_getter = getattr(self._devmgr, get_method)
            template_id = id or hdt['hosting_device_template']['id']
            driver = driver_getter(context, template_id)
            if test_for_none:
                self.assertIsNone(driver)
            else:
                self.assertIsNotNone(driver)

    # driver request tests
    def test_get_hosting_device_driver(self):
        self._test_get_driver('get_hosting_device_driver')

    def test_get_non_existent_hosting_device_driver_returns_none(self):
        self._test_get_driver('get_hosting_device_driver', 'bogus_id', True)

    def test_get_plugging_device_driver(self):
        self._test_get_driver('get_hosting_device_plugging_driver')

    def test_get_non_existent_plugging_device_driver_returns_none(self):
        self._test_get_driver('get_hosting_device_plugging_driver', 'bogus_id',
                              True)

    # get device info tests
    def test_get_device_info_for_agent(self):
        device_id = "device_XYZ"
        with self.hosting_device_template() as hdt, self.port(
                subnet=self._mgmt_subnet) as mgmt_port:
            creds = device_manager_test_support._uuid()
            mgmt_ip = mgmt_port['port']['fixed_ips'][0]['ip_address']
            with self.hosting_device(
                    device_id=device_id,
                    template_id=hdt['hosting_device_template']['id'],
                    management_port_id=mgmt_port['port']['id'],
                    credentials_id=creds) as hd:
                context = self._get_test_context(
                    tenant_id=hdt['hosting_device_template']['tenant_id'],
                    is_admin=True)
                hd_id = hd['hosting_device']['id']
                hd_db = self._devmgr._get_hosting_device(context, hd_id)
                info = self._devmgr.get_device_info_for_agent(context, hd_db)
                self.assertEqual(mgmt_ip, info['management_ip_address'])

    def test_get_device_info_for_agent_no_mgmt_port(self):
        device_id = "device_XYZ"
        with self.hosting_device_template() as hdt:
            creds = device_manager_test_support._uuid()
            mgmt_ip = '192.168.0.55'
            with self.hosting_device(
                    device_id=device_id,
                    template_id=hdt['hosting_device_template']['id'],
                    management_ip_address=mgmt_ip,
                    management_port_id=None,
                    credentials_id=creds) as hd:
                context = self._get_test_context(
                    tenant_id=hdt['hosting_device_template']['tenant_id'],
                    is_admin=True)
                hd_id = hd['hosting_device']['id']
                hd_db = self._devmgr._get_hosting_device(context, hd_id)
                info = self._devmgr.get_device_info_for_agent(context, hd_db)
                self.assertEqual(mgmt_ip, info['management_ip_address'])

    def _set_ownership(self, bound_status, tenant_id, other_tenant_id=None):
        if bound_status == UNBOUND:
            return None
        elif bound_status == OTHER:
            return other_tenant_id or self._other_tenant_id
        else:
            return tenant_id

    # slot allocation and release test helper:
    # succeeds means returns True, fails means returns False
    def _test_slots(self, expected_result=True, expected_bind=UNBOUND,
                    expected_allocation=VM_SLOT_CAPACITY,
                    num_requested=VM_SLOT_CAPACITY,
                    slot_capacity=VM_SLOT_CAPACITY, initial_bind=UNBOUND,
                    bind=False, auto_delete=True, is_admin=False,
                    pool_maintenance_expected=True, test_release=False,
                    expected_release_result=True, expected_final_allocation=0,
                    expected_release_bind=UNBOUND,
                    num_to_release=VM_SLOT_CAPACITY,
                    release_pool_maintenance_expected=True):
        with self.hosting_device_template(
                slot_capacity=slot_capacity) as hdt:
            with self.port(subnet=self._mgmt_subnet) as mgmt_port:
                resource = self._get_fake_resource()
                tenant_bound = self._set_ownership(
                    initial_bind, resource['tenant_id'])
                with self.hosting_device(
                        template_id=hdt['hosting_device_template']['id'],
                        management_port_id=mgmt_port['port']['id'],
                        tenant_bound=tenant_bound,
                        auto_delete=auto_delete) as hd:
                    context = self._get_test_context(
                        tenant_id=hdt['hosting_device_template']['tenant_id'],
                        is_admin=is_admin)
                    hd_db = self._devmgr._get_hosting_device(
                        context, hd['hosting_device']['id'])
                    with mock.patch.object(
                            hdm_db.HostingDeviceManagerMixin,
                            '_dispatch_pool_maintenance_job') as pm_mock:
                        result = self._devmgr.acquire_hosting_device_slots(
                            context, hd_db, resource, 'router', L3_ROUTER_NAT,
                            num_requested, bind)
                        allocation = self._devmgr.get_slot_allocation(
                            context, resource_id=resource['id'])
                        self.assertEqual(expected_result, result)
                        self.assertEqual(expected_allocation, allocation)
                        expected_bind = self._set_ownership(
                            expected_bind, resource['tenant_id'])
                        self.assertEqual(expected_bind, hd_db.tenant_bound)
                        if pool_maintenance_expected:
                            pm_mock.assert_called_once_with(mock.ANY)
                            num_calls = 1
                        else:
                            pm_mock.assert_not_called()
                            num_calls = 0
                        if test_release:
                            result = self._devmgr.release_hosting_device_slots(
                                context, hd_db, resource, num_to_release)
                            if not test_release:
                                return
                            allocation = self._devmgr.get_slot_allocation(
                                context, resource_id=resource['id'])
                            self.assertEqual(expected_release_result, result)
                            self.assertEqual(expected_final_allocation,
                                             allocation)
                            expected_release_bind = self._set_ownership(
                                expected_release_bind, resource['tenant_id'])
                            self.assertEqual(expected_release_bind,
                                             hd_db.tenant_bound)
                            if release_pool_maintenance_expected:
                                num_calls += 1
                            self.assertEqual(num_calls, pm_mock.call_count)
                        else:
                            # ensure we clean up everything
                            num_to_release = 0
                        to_clean_up = num_requested - num_to_release
                        if to_clean_up < 0:
                            to_clean_up = num_requested
                        if to_clean_up:
                            self._devmgr.release_hosting_device_slots(
                                context, hd_db, resource, to_clean_up)

    # slot allocation tests
    def test_acquire_with_slot_surplus_in_owned_hosting_device_succeeds(self):
        self._test_slots(expected_bind=REQUESTER, initial_bind=REQUESTER,
                         bind=True)

    def test_acquire_with_slot_surplus_in_shared_hosting_device_succeeds(self):
        self._test_slots()

    def test_acquire_with_slot_surplus_take_hosting_device_ownership_succeeds(
            self):
        self._test_slots(expected_bind=REQUESTER, initial_bind=UNBOUND,
                         bind=True)

    def test_acquire_with_slot_surplus_drop_hosting_device_ownership_succeeds(
            self):
        self._test_slots(expected_bind=UNBOUND, initial_bind=REQUESTER,
                         bind=False)

    def test_acquire_slots_release_hosting_device_ownership_affects_all(self):
        #TODO(bobmel): Implement this unit test
        pass

    def test_acquire_slots_in_other_owned_hosting_device_fails(self):
        self._test_slots(expected_result=False, expected_bind=OTHER,
                         expected_allocation=0, initial_bind=OTHER,
                         pool_maintenance_expected=False)

    def test_acquire_slots_take_ownership_of_other_owned_hosting_device_fails(
            self):
        self._test_slots(expected_result=False, expected_bind=OTHER,
                         expected_allocation=0, initial_bind=OTHER,
                         bind=True, pool_maintenance_expected=False)

    def test_acquire_slots_take_ownership_of_multi_tenant_hosting_device_fails(
            self):
        #TODO(bobmel): Implement this unit test
        pass

    def test_acquire_with_slot_deficit_in_owned_hosting_device_fails(self):
        self._test_slots(expected_result=False, expected_bind=REQUESTER,
                         expected_allocation=0, initial_bind=REQUESTER,
                         num_requested=VM_SLOT_CAPACITY + 1)

    def test_acquire_with_slot_deficit_in_shared_hosting_device_fails(self):
        self._test_slots(expected_result=False, expected_bind=UNBOUND,
                         expected_allocation=0,
                         num_requested=VM_SLOT_CAPACITY + 1)

    def test_acquire_with_slot_deficit_in_other_owned_hosting_device_fails(
            self):
        self._test_slots(expected_result=False, expected_bind=OTHER,
                         expected_allocation=0, initial_bind=OTHER,
                         num_requested=VM_SLOT_CAPACITY + 1,
                         pool_maintenance_expected=False)

    # slot release tests
    def test_release_allocated_slots_in_owned_hosting_device_succeeds(self):
        self._test_slots(expected_bind=REQUESTER, initial_bind=REQUESTER,
                         bind=True, test_release=True,
                         expected_release_bind=REQUESTER,
                         expected_final_allocation=1,
                         num_to_release=VM_SLOT_CAPACITY - 1)

    def test_release_allocated_slots_in_shared_hosting_device_succeeds(self):
        self._test_slots(test_release=True, expected_final_allocation=1,
                         num_to_release=VM_SLOT_CAPACITY - 1)

    def test_release_all_slots_returns_hosting_device_ownership(self):
        self._test_slots(expected_bind=REQUESTER, initial_bind=REQUESTER,
                         bind=True, test_release=True,
                         expected_release_bind=UNBOUND)

    def test_release_slots_in_other_owned_hosting_device_fails(self):
        self._test_slots(expected_result=False, expected_bind=OTHER,
                         expected_allocation=0, initial_bind=OTHER,
                         pool_maintenance_expected=False,
                         test_release=True, expected_release_result=False,
                         expected_release_bind=OTHER,
                         expected_final_allocation=0,
                         num_to_release=VM_SLOT_CAPACITY - 1,
                         release_pool_maintenance_expected=False)

    def test_release_too_many_slots_in_owned_hosting_device_fails(self):
        self._test_slots(expected_bind=REQUESTER, initial_bind=REQUESTER,
                         bind=True, test_release=True,
                         expected_release_result=False,
                         expected_release_bind=REQUESTER,
                         expected_final_allocation=VM_SLOT_CAPACITY,
                         num_to_release=VM_SLOT_CAPACITY + 1)

    def test_release_too_many_slots_in_shared_hosting_device_fails(self):
        self._test_slots(test_release=True, expected_release_result=False,
                         expected_release_bind=UNBOUND,
                         expected_final_allocation=VM_SLOT_CAPACITY,
                         num_to_release=VM_SLOT_CAPACITY + 1)

    def test_release_too_many_slots_in_other_owned_hosting_device_fails(
            self):
        self._test_slots(expected_result=False, expected_bind=OTHER,
                         expected_allocation=0, initial_bind=OTHER,
                         pool_maintenance_expected=False,
                         test_release=True, expected_release_result=False,
                         expected_release_bind=OTHER,
                         expected_final_allocation=0,
                         num_to_release=VM_SLOT_CAPACITY + 1,
                         release_pool_maintenance_expected=False)

    def test_release_all_slots_by_negative_num_argument_shared_hosting_device(
            self):
        self._test_slots(test_release=True, expected_final_allocation=0,
                         num_to_release=-1)

    def test_release_all_slots_by_negative_num_argument_owned_hosting_device(
            self):
        self._test_slots(expected_bind=REQUESTER, initial_bind=REQUESTER,
                 bind=True, test_release=True, expected_release_bind=UNBOUND,
                 expected_final_allocation=0, num_to_release=-1)

    # hosting device deletion test helper
    def _test_delete(self, to_delete=None, auto_delete=None, no_delete=None,
                     force_delete=True, expected_num_remaining=0):
        auto_delete = auto_delete or [True, False, False, True, True]
        no_delete = no_delete or [True, True, True, True, True]
        with self.hosting_device_template() as hdt1,\
                self.hosting_device_template() as hdt2:
            hdt0_id = hdt1['hosting_device_template']['id']
            hdt1_id = hdt2['hosting_device_template']['id']
            with self.port(subnet=self._mgmt_subnet,
                           no_delete=no_delete[0]) as mgmt_port0,\
                    self.port(subnet=self._mgmt_subnet,
                              no_delete=no_delete[1]) as mgmt_port1,\
                    self.port(subnet=self._mgmt_subnet,
                              no_delete=no_delete[2]) as mgmt_port2,\
                    self.port(subnet=self._mgmt_subnet,
                              no_delete=no_delete[3]) as mgmt_port3,\
                    self.port(subnet=self._mgmt_subnet,
                              no_delete=no_delete[4]) as mgmt_port4:
                mp0_id = mgmt_port0['port']['id']
                mp1_id = mgmt_port1['port']['id']
                mp2_id = mgmt_port2['port']['id']
                mp3_id = mgmt_port3['port']['id']
                mp4_id = mgmt_port4['port']['id']
                with self.hosting_device(
                        device_id='0_hdt0_id', template_id=hdt0_id,
                        management_port_id=mp0_id, auto_delete=auto_delete[0],
                        no_delete=no_delete[0]),\
                    self.hosting_device(
                        device_id='1_hdt1_id', template_id=hdt1_id,
                        management_port_id=mp1_id, auto_delete=auto_delete[1],
                        no_delete=no_delete[1]),\
                    self.hosting_device(
                        device_id='2_hdt0_id', template_id=hdt0_id,
                        management_port_id=mp2_id, auto_delete=auto_delete[2],
                        no_delete=no_delete[2]),\
                    self.hosting_device(
                        device_id='3_hdt0_id', template_id=hdt0_id,
                        management_port_id=mp3_id,
                        auto_delete=auto_delete[3],
                        no_delete=no_delete[3]),\
                    self.hosting_device(
                        device_id='4_hdt1_id', template_id=hdt1_id,
                        management_port_id=mp4_id, auto_delete=auto_delete[4],
                        no_delete=no_delete[4]):
                        context = self._get_test_context(is_admin=True)
                        if to_delete is None:
                            self._devmgr.delete_all_hosting_devices(
                                context, force_delete)
                        elif to_delete == 0:
                            template = (
                                self._devmgr._get_hosting_device_template(
                                    context, hdt0_id))
                            (self._devmgr.
                                delete_all_hosting_devices_by_template(
                                    context, template, force_delete))
                        else:
                            template = (
                                self._devmgr._get_hosting_device_template(
                                    context, hdt1_id))
                            (self._devmgr.
                                delete_all_hosting_devices_by_template(
                                    context, template, force_delete))
                        result_hds = self._list(
                            'hosting_devices')['hosting_devices']
                        self.assertEqual(expected_num_remaining,
                                         len(result_hds))

    # hosting device deletion tests
    def test_delete_all_hosting_devices(self):
        self._test_delete()

    def test_delete_all_managed_hosting_devices(self):
        self._test_delete(no_delete=[True, False, False, True, True],
                          force_delete=False, expected_num_remaining=2)

    def test_delete_all_hosting_devices_by_template(self):
        self._test_delete(to_delete=1, expected_num_remaining=3,
                          no_delete=[False, True, False, False, True])

    def test_delete_all_managed_hosting_devices_by_template(self):
        self._test_delete(to_delete=1, expected_num_remaining=4,
                          no_delete=[False, False, False, False, True],
                          force_delete=False)

    # handled failed hosting device test helper
    def _test_failed_hosting_device(self, host_category=VM_CATEGORY,
                                    expected_num_remaining=0,
                                    auto_delete=True, no_delete=True):
        with self.hosting_device_template(host_category=host_category) as hdt:
            hdt_id = hdt['hosting_device_template']['id']
            with self.port(subnet=self._mgmt_subnet,
                           no_delete=no_delete) as mgmt_port:
                with self.hosting_device(
                        template_id=hdt_id,
                        management_port_id=mgmt_port['port']['id'],
                        auto_delete=auto_delete, no_delete=no_delete) as hd:
                    with mock.patch('networking_cisco.backwards_compatibility.'
                                    'get_plugin'):
                        hd_id = hd['hosting_device']['id']
                        m2 = mock.MagicMock()
                        self._devmgr.agent_notifiers = {
                            c_constants.AGENT_TYPE_CFG: m2}
                        context = self._get_test_context()
                        self._devmgr.handle_non_responding_hosting_devices(
                            context, None, [hd_id])
                        result_hds = self._list('hosting_devices')[
                            'hosting_devices']
                        self.assertEqual(expected_num_remaining,
                                         len(result_hds))
                        l3mock = (bc.get_plugin().
                                  handle_non_responding_hosting_devices)
                        l3mock.assert_called_once_with(mock.ANY, mock.ANY,
                                                       {hd_id: {}})
                        if expected_num_remaining == 0:
                            m2.hosting_devices_removed.assert_called_once_with(
                                mock.ANY, {hd_id: {}}, False, None)

    # handled failed hosting device tests
    def test_failed_managed_vm_based_hosting_device_gets_deleted(self):
        self._test_failed_hosting_device()

    def test_failed_non_managed_vm_based_hosting_device_not_deleted(self):
        self._test_failed_hosting_device(expected_num_remaining=1,
                                         auto_delete=False, no_delete=False)

    def test_failed_non_vm_based_hosting_device_not_deleted(self):
        self._test_failed_hosting_device(host_category=HW_CATEGORY,
                                         expected_num_remaining=1,
                                         no_delete=False)

    # hosting device pool maintenance test helper
    def _test_pool_maintenance(self, desired_slots_free=10, slot_capacity=3,
                               host_category=VM_CATEGORY, expected=15,
                               define_credentials=True):
        with self.hosting_device_template(
                host_category=host_category, slot_capacity=slot_capacity,
                desired_slots_free=desired_slots_free,
                plugging_driver=TEST_PLUGGING_DRIVER) as hdt:
            hdt_id = hdt['hosting_device_template']['id']
            creds_id = (DEFAULT_CREDENTIALS_ID if define_credentials is True
                        else 'non_existent_id')
            credentials = {'user_name': 'bob', 'password': 'tooEasy'}
            with mock.patch.dict(
                    self.plugin._credentials,
                    {creds_id: credentials}),\
                    self.port(subnet=self._mgmt_subnet,
                              no_delete=True) as mgmt_port1,\
                    self.port(subnet=self._mgmt_subnet,
                              no_delete=True) as mgmt_port2:
                with self.hosting_device(
                        template_id=hdt_id,
                        management_port_id=mgmt_port1['port']['id'],
                        auto_delete=True, no_delete=True),\
                     self.hosting_device(
                         template_id=hdt_id,
                         management_port_id=mgmt_port2['port']['id'],
                         auto_delete=True, no_delete=True):
                    context = self._get_test_context(is_admin=True)
                    template = self._devmgr._get_hosting_device_template(
                        context, hdt_id)

                    self._devmgr._gt_pool = mock.MagicMock()
                    self._devmgr._gt_pool.spawn_n.side_effect = (
                        lambda fcn, ctx, tmplt: fcn(ctx, tmplt))
                    self._devmgr._dispatch_pool_maintenance_job(
                        template)
                    result_hds = self._list(
                        'hosting_devices')['hosting_devices']
                    self.assertEqual(expected, len(result_hds) * slot_capacity)
            self._devmgr.delete_all_hosting_devices(context, True)

    # hosting device pool maintenance tests
    def test_vm_based_hosting_device_excessive_slot_deficit_adds_slots(self):
        self._test_pool_maintenance()

    def test_vm_based_hosting_device_excessive_slot_deficit_no_credentials(
            self):
        # no slots are added if credentials for template are missing
        self._test_pool_maintenance(expected=6, define_credentials=False)

    def test_vm_based_hosting_device_marginal_slot_deficit_no_change(self):
        self._test_pool_maintenance(desired_slots_free=7, expected=6)

    def test_vm_based_hosting_device_excessive_slot_surplus_removes_slots(
            self):
        self._test_pool_maintenance(desired_slots_free=3, expected=3)

    def test_vm_based_hosting_device_marginal_slot_surplus_no_change(self):
        self._test_pool_maintenance(desired_slots_free=5, expected=6)

    def test_hw_based_hosting_device_no_change(self):
        self._test_pool_maintenance(host_category=HW_CATEGORY, expected=6)



# Copyright 2015 Cisco Systems, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#


"""
This file provides a wrapper to novaclient API, for getting the instacne's
information such as display_name.
"""
from keystoneauth1.identity import generic
from keystoneauth1 import session
from novaclient import api_versions
from novaclient import client as nova_client
from novaclient import exceptions as nexc

from networking_cisco._i18n import _LE

from networking_cisco.apps.saf.common import config
from networking_cisco.apps.saf.common import dfa_logger as logging


LOG = logging.getLogger(__name__)


class DFAInstanceAPI(object):

    """This class provides API to get information for a given instance."""

    def __init__(self):
        self._cfg = config.CiscoDFAConfig().cfg
        self._inst_info_cache = {}
        user = self._cfg.nova.username
        project = self._cfg.nova.project_name
        passwd = self._cfg.nova.password
        url = self._cfg.nova.auth_url
        region = self._cfg.nova.region_name
        u_domain = self._cfg.nova.user_domain_name
        p_domain = self._cfg.nova.project_domain_name
        api_v = self._cfg.nova.api_version
        auth = generic.Password(auth_url=url,
                                username=user,
                                password=passwd,
                                project_name=project,
                                project_domain_name=p_domain,
                                user_domain_name=u_domain)
        sess = session.Session(auth=auth)

        self._novaclnt = nova_client.Client(api_versions.APIVersion(api_v),
                                     session=sess,
                                     region_name=region)
        LOG.debug('DFAInstanceAPI: initialization done...')

    def _get_instances_for_project(self, project_id):
        """Return all instances for a given project.

        :project_id: UUID of project (tenant)
        """
        search_opts = {'marker': None,
                       'all_tenants': True,
                       'project_id': project_id}
        try:
            servers = self._novaclnt.servers.list(True, search_opts)
            LOG.debug('_get_instances_for_project: servers=%s', servers)
            return servers
        except nexc.Unauthorized:
            emsg = (_LE('Failed to get novaclient:Unauthorised '
                    'project_id=%(proj)s user=%(user)s'),
                    {'proj': self._project_id, 'user': self._user_name})
            LOG.exception(emsg)
            raise nexc.ClientException(emsg)
        except nexc.AuthorizationFailure as err:
            emsg = (_LE("Failed to get novaclient %s"))
            LOG.exception(emsg, err)
            raise nexc.ClientException(emsg % err)

    def get_instance_for_uuid(self, uuid, project_id):
        """Return instance name for given uuid of an instance and project.

        :uuid: Instance's UUID
        :project_id: UUID of project (tenant)
        """
        instance_name = self._inst_info_cache.get((uuid, project_id))
        if instance_name:
            return instance_name
        instances = self._get_instances_for_project(project_id)
        for inst in instances:
            if inst.id.replace('-', '') == uuid:
                LOG.debug('get_instance_for_uuid: name=%s', inst.name)
                instance_name = inst.name
                self._inst_info_cache[(uuid, project_id)] = instance_name
                return instance_name
        return instance_name



# Copyright 2015 Cisco Systems, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#


"""This module provides APIs for communicating with DCNM."""


import re
import requests
import sys

from oslo_serialization import jsonutils

from networking_cisco.apps.saf.common import dfa_exceptions as dexc
from networking_cisco.apps.saf.common import dfa_logger as logging


LOG = logging.getLogger(__name__)
UNKNOWN_SRVN_NODE_IP = '0.0.0.0'
UNKNOWN_DCI_ID = -1


class DFARESTClient(object):

    """DFA client class that provides APIs to interact with DCNM."""

    def __init__(self, cfg):
        self._base_ver = '7.1(0)'
        self._is_iplus = False
        self._ip = cfg.dcnm.dcnm_ip
        self._user = cfg.dcnm.dcnm_user
        self._pwd = cfg.dcnm.dcnm_password
        self._part_name = cfg.dcnm.default_partition_name
        if (not self._ip) or (not self._user) or (not self._pwd):
            msg = ("[DFARESTClient] Input DCNM IP, user name or password"
                   "parameter is not specified")
            raise ValueError(msg)

        self._req_headers = {'Accept': 'application/json',
                             'Content-Type': 'application/json; charset=UTF-8'}

        self.default_cfg_profile = cfg.dcnm.default_cfg_profile
        self.default_vrf_profile = cfg.dcnm.default_vrf_profile
        # url timeout: 10 seconds
        self.timeout_resp = (10 if not cfg.dcnm.timeout_resp else
                             cfg.dcnm.timeout_resp)
        self._exp_time = 100000
        self._resp_ok = (requests.codes.ok, requests.codes.created,
                         requests.codes.accepted)

        self.dcnm_protocol = self.get_dcnm_protocol()
        # Fill the urls for DCNM Rest API's.
        self.fill_urls()

        self._cur_ver = self.get_version()
        self._detect_iplus()

        # Update the default network profile based on version of DCNM.
        self._set_default_cfg_profile()
        self._default_md = None

    def _detect_iplus(self):
        """Check the DCNM version and determine if it's for iplus"""

        ver_expr = "([0-9]+)\.([0-9]+)\((.*)\)"
        re.compile(ver_expr)
        v1 = re.match(ver_expr, self._cur_ver)
        v2 = re.match(ver_expr, self._base_ver)

        if int(v1.group(1)) > int(v2.group(1)):
            self._is_iplus = True
        elif int(v1.group(1)) == int(v2.group(1)):
            if int(v1.group(2)) > int(v2.group(2)):
                self._is_iplus = True
            elif int(v1.group(2)) == int(v2.group(2)):
                self._is_iplus = v1.group(3) >= v2.group(3)

        LOG.info("DCNM version: %(cur_ver)s, iplus: %(is_iplus)s",
                 {'cur_ver': self._cur_ver, 'is_iplus': self._is_iplus})

    def _failure_msg(self, response):
        return "[%s] %s" % (response.status_code, response.text)

    def get_segmentid_range(self, orchestrator_id):
        """Get segment id range from DCNM. """

        url = "%s/%s" % (self._segmentid_ranges_url, orchestrator_id)

        res = self._send_request('GET', url, None, 'segment-id range')
        if res and res.status_code in self._resp_ok:
            return res.json()

    def set_segmentid_range(self, orchestrator_id, segid_min, segid_max):
        """set segment id range in DCNM. """

        url = self._segmentid_ranges_url

        payload = {'orchestratorId': orchestrator_id,
                   'segmentIdRanges': "%s-%s" % (segid_min, segid_max)}

        res = self._send_request('POST', url, payload, 'segment-id range')
        if not (res and res.status_code in self._resp_ok):
            LOG.error("Failed to set segment id range for orchestrator "
                      "%(orch)s on DCNM: %(text)s",
                      {'orch': orchestrator_id, 'text': res.text})
            raise dexc.DfaClientRequestFailed(reason=self._failure_msg(res))

    def update_segmentid_range(self, orchestrator_id, segid_min, segid_max):
        """update segment id range in DCNM. """
        url = "%s/%s" % (self._segmentid_ranges_url, orchestrator_id)

        payload = {'orchestratorId': orchestrator_id,
                   'segmentIdRanges': "%s-%s" % (segid_min, segid_max)}

        res = self._send_request('PUT', url, payload, 'segment-id range')
        if not (res and res.status_code in self._resp_ok):
            LOG.error("Failed to update segment id range for orchestrator "
                      "%(orch)s on DCNM: %(text)s",
                      {'orch': orchestrator_id, 'text': res.text})
            raise dexc.DfaClientRequestFailed(reason=self._failure_msg(res))

    def _set_default_cfg_profile(self):
        """Set default network config profile.

        Check whether the default_cfg_profile value exist in the current
        version of DCNM. If not, set it to new default value which is supported
        by latest version.
        """
        try:
            cfgplist = self.config_profile_list()
            if self.default_cfg_profile not in cfgplist:
                self.default_cfg_profile = ('defaultNetworkUniversalEfProfile'
                                            if self._is_iplus else
                                            'defaultNetworkIpv4EfProfile')
        except dexc.DfaClientRequestFailed:
            LOG.error("Failed to send request to DCNM.")
            self.default_cfg_profile = 'defaultNetworkIpv4EfProfile'

    def _create_network(self, network_info):
        """Send create network request to DCNM.

        :param network_info: network parameters to be created on DCNM
        """
        url = self._create_network_url % (network_info['organizationName'],
                                          network_info['partitionName'])
        payload = network_info

        LOG.info('url %(url)s payload %(payload)s',
                 {'url': url, 'payload': payload})
        return self._send_request('POST', url, payload, 'network')

    def _config_profile_get(self, thisprofile):
        """Get information of a config profile from DCNM.

        :param thisprofile: network config profile in request
        """
        url = self._cfg_profile_get_url % (thisprofile)
        payload = {}

        res = self._send_request('GET', url, payload, 'config-profile')
        if res and res.status_code in self._resp_ok:
            return res.json()

    def _config_profile_list(self):
        """Get list of supported config profile from DCNM."""
        url = self._cfg_profile_list_url
        payload = {}

        try:
            res = self._send_request('GET', url, payload, 'config-profile')
            if res and res.status_code in self._resp_ok:
                return res.json()
        except dexc.DfaClientRequestFailed:
            LOG.error("Failed to send request to DCNM.")

    def _get_settings(self):
        """Get global mobility domain from DCNM."""
        url = self._global_settings_url
        payload = {}
        res = self._send_request('GET', url, payload, 'settings')
        if res and res.status_code in self._resp_ok:
            return res.json()

    def _set_default_mobility_domain(self):
        settings = self._get_settings()
        LOG.info("settings is %s" % settings)

        if ('globalMobilityDomain' in settings.keys()):
            global_md = settings.get('globalMobilityDomain')
            self._default_md = global_md.get('name')
            LOG.info("setting default md to be %s" % self._default_md)
        else:
            self._default_md = "md0"

    def _create_org(self, orch_id, name, desc):
        """Create organization on the DCNM.

        :param orch_id: orchestrator ID
        :param name: Name of organization
        :param desc: Description of organization
        """
        url = self._org_url
        payload = {
            "organizationName": name,
            "description": name if len(desc) == 0 else desc,
            "orchestrationSource": orch_id}

        return self._send_request('POST', url, payload, 'organization')

    def _create_or_update_partition(self, org_name, part_name, desc,
                                    dci_id=UNKNOWN_DCI_ID, vrf_prof=None,
                                    service_node_ip=UNKNOWN_SRVN_NODE_IP,
                                    operation='POST'):
        """Send create or update partition request to the DCNM.

        :param org_name: name of organization
        :param part_name: name of partition
        :param desc: description of partition
        :dci_id: DCI ID for inter-DC
        :vrf_prof: VRF Profile Name
        :service_node_ip: Service Node's Address
        """
        if part_name is None:
            part_name = self._part_name
        if vrf_prof is None or dci_id == UNKNOWN_DCI_ID or (
           service_node_ip == UNKNOWN_SRVN_NODE_IP):
            part_info = self._get_partition(org_name, part_name)
        if vrf_prof is None:
            vrf_prof = self.get_partition_vrfProf(org_name, part_name,
                                                  part_info=part_info)
        if dci_id == UNKNOWN_DCI_ID:
            dci_id = self.get_partition_dciId(org_name, part_name,
                                              part_info=part_info)
        if service_node_ip == UNKNOWN_SRVN_NODE_IP:
            service_node_ip = self.get_partition_serviceNodeIp(
                org_name, part_name, part_info=part_info)
        url = ((self._create_part_url % (org_name)) if operation == 'POST' else
               self._update_part_url % (org_name, part_name))

        payload = {
            "partitionName": part_name,
            "description": part_name if len(desc) == 0 else desc,
            "serviceNodeIpAddress": service_node_ip,
            "organizationName": org_name}

        # Check the DCNM version and find out whether it is need to have
        # extra payload for the new version when creating/updating a partition.
        if self._is_iplus:
            # Need to add extra payload for the new version.
            enable_dci = "true" if dci_id and int(dci_id) != 0 else "false"
            extra_payload = {
                "vrfProfileName": vrf_prof,
                "vrfName": ':'.join((org_name, part_name)),
                "dciId": dci_id,
                "enableDCIExtension": enable_dci}
            payload.update(extra_payload)

        return self._send_request(operation, url, payload, 'partition')

    def _get_partition(self, org_name, part_name=None):
        """send get partition request to the DCNM.

        :param org_name: name of organization
        :param part_name: name of partition
        """
        if part_name is None:
            part_name = self._part_name
        url = self._update_part_url % (org_name, part_name)
        res = self._send_request("GET", url, '', 'partition')
        if res and res.status_code in self._resp_ok:
            return res.json()

    def update_partition_static_route(self, org_name, part_name,
                                      static_ip_list, vrf_prof=None,
                                      service_node_ip=None):
        """Send static route update requests to DCNM.

        :param org_name: name of organization
        :param part_name: name of partition
        :static_ip_list: List of static IP addresses
        :vrf_prof: VRF Profile
        :service_node_ip: Service Node IP address
        """
        if part_name is None:
            part_name = self._part_name
        if vrf_prof is None:
            vrf_prof = self.default_vrf_profile
        operation = 'PUT'
        url = (self._update_part_url % (org_name, part_name))
        ip_str = ''
        ip_cnt = 0
        for ip in static_ip_list:
            ip_sub = "$n0" + str(ip_cnt) + "=" + str(ip) + ";"
            ip_str = ip_str + ip_sub
            ip_cnt = ip_cnt + 1
        cfg_args = {
            "$vrfName=" + org_name + ':' + part_name + ";"
            "$include_serviceNodeIpAddress=" + service_node_ip + ";"
            + ip_str
        }
        cfg_args = ';'.join(cfg_args)
        payload = {
            "partitionName": part_name,
            "organizationName": org_name,
            "dciExtensionStatus": "Not configured",
            "vrfProfileName": vrf_prof,
            "vrfName": ':'.join((org_name, part_name)),
            "configArg": cfg_args}

        res = self._send_request(operation, url, payload, 'partition')
        return (res is not None and res.status_code in self._resp_ok)

    def _delete_org(self, org_name):
        """Send organization delete request to DCNM.

        :param org_name: name of organization to be deleted
        """
        url = self._del_org_url % (org_name)
        return self._send_request('DELETE', url, '', 'organization')

    def _delete_partition(self, org_name, partition_name):
        """Send partition delete request to DCNM.

        :param org_name: name of organization
        :param partition_name: name of partition
        """
        url = self._del_part % (org_name, partition_name)
        return self._send_request('DELETE', url, '', 'partition')

    def _delete_network(self, network_info):
        """Send network delete request to DCNM.

        :param network_info: contains network info to be deleted.
        """
        org_name = network_info.get('organizationName', '')
        part_name = network_info.get('partitionName', '')
        segment_id = network_info['segmentId']
        if 'mobDomainName' in network_info:
            vlan_id = network_info['vlanId']
            mob_dom_name = network_info['mobDomainName']
            url = self._network_mob_url % (org_name, part_name, vlan_id,
                                           mob_dom_name)
        else:
            url = self._network_url % (org_name, part_name, segment_id)
        return self._send_request('DELETE', url, '', 'network')

    def _get_network(self, network_info):
        """Send network get request to DCNM.

        :param network_info: contains network info to query.
        """
        org_name = network_info.get('organizationName', '')
        part_name = network_info.get('partitionName', '')
        segment_id = network_info['segmentId']
        url = self._network_url % (org_name, part_name, segment_id)
        return self._send_request('GET', url, '', 'network')

    def _login_request(self, url_login):
        """Internal function to send login request. """

        expiration_time = self._exp_time
        payload = {'expirationTime': expiration_time}
        # TODO(padkrish), after testing with certificates, make the
        # verify option configurable.
        res = requests.post(url_login,
                            data=jsonutils.dumps(payload),
                            headers=self._req_headers,
                            auth=(self._user, self._pwd),
                            timeout=self.timeout_resp, verify=False)
        session_id = ''
        if res and res.status_code in self._resp_ok:
            session_id = res.json().get('Dcnm-Token')
        self._req_headers.update({'Dcnm-Token': session_id})

    def _login(self):
        """Login request to DCNM. """

        self._login_request(self._login_url)

    def _logout_request(self, url_logout):
        """Internal logout request to DCNM. """

        requests.post(url_logout,
                      headers=self._req_headers,
                      timeout=self.timeout_resp, verify=False)

    def _logout(self, url_arg=None):
        """Logout request to DCNM."""

        self._logout_request(self._logout_url)

    def _send_request(self, operation, url, payload, desc):
        """Send request to DCNM."""

        res = None
        try:
            payload_json = None
            if payload and payload != '':
                payload_json = jsonutils.dumps(payload)
            self._login()
            desc_lookup = {'POST': ' creation', 'PUT': ' update',
                           'DELETE': ' deletion', 'GET': ' get'}

            res = requests.request(operation, url, data=payload_json,
                                   headers=self._req_headers,
                                   timeout=self.timeout_resp, verify=False)
            desc += desc_lookup.get(operation, operation.lower())
            LOG.info("DCNM-send_request: %(desc)s %(url)s %(pld)s",
                     {'desc': desc, 'url': url, 'pld': payload})

            self._logout()
        except (requests.HTTPError, requests.Timeout,
                requests.ConnectionError) as exc:
            LOG.exception('Error during request: %s', exc)
            raise dexc.DfaClientRequestFailed(reason=exc)

        return res

    def config_profile_list(self):
        """Return config profile list from DCNM."""

        these_profiles = self._config_profile_list() or []
        profile_list = [q for p in these_profiles for q in
                        [p.get('profileName')]]
        return profile_list

    def config_profile_fwding_mode_get(self, profile_name):
        """Return forwarding mode of given config profile."""

        profile_params = self._config_profile_get(profile_name)
        fwd_cli = 'fabric forwarding mode proxy-gateway'
        if profile_params and fwd_cli in profile_params['configCommands']:
            return 'proxy-gateway'
        else:
            return 'anycast-gateway'

    def get_config_profile_for_network(self, net_name):
        """Get the list of profiles."""

        cfgplist = self.config_profile_list()
        cfgname = net_name.partition(':')[2]

        cfgtuple = set()
        for cfg_prof in cfgplist:
            if cfg_prof.startswith('defaultNetwork'):
                cfg_alias = (cfg_prof.split('defaultNetwork')[1].
                             split('Profile')[0])
            elif cfg_prof.endswith('Profile'):
                cfg_alias = cfg_prof.split('Profile')[0]
            else:
                cfg_alias = cfg_prof
            cfgtuple.update([(cfg_prof, cfg_alias)])
        cfgp = [a for a, b in cfgtuple if cfgname == b]
        prof = cfgp[0] if cfgp else self.default_cfg_profile
        fwd_mod = self.config_profile_fwding_mode_get(prof)
        return (prof, fwd_mod)

    def create_network(self, tenant_name, network, subnet,
                       dhcp_range=True):
        """Create network on the DCNM.

        :param tenant_name: name of tenant the network belongs to
        :param network: network parameters
        :param subnet: subnet parameters of the network
        """
        seg_id = str(network.segmentation_id)
        subnet_ip_mask = subnet.cidr.split('/')
        gw_ip = subnet.gateway_ip
        cfg_args = [
            "$segmentId=" + seg_id,
            "$netMaskLength=" + subnet_ip_mask[1],
            "$gatewayIpAddress=" + gw_ip,
            "$networkName=" + network.name,
            "$vlanId=0",
            "$vrfName=" + tenant_name + ':' + self._part_name
        ]
        cfg_args = ';'.join(cfg_args)

        ip_range = ','.join(["%s-%s" % (p['start'], p['end']) for p in
                             subnet.allocation_pools])

        dhcp_scopes = {'ipRange': ip_range,
                       'subnet': subnet.cidr,
                       'gateway': gw_ip}

        network_info = {"segmentId": seg_id,
                        "vlanId": "0",
                        "mobilityDomainId": "None",
                        "profileName": network.config_profile,
                        "networkName": network.name,
                        "configArg": cfg_args,
                        "organizationName": tenant_name,
                        "partitionName": self._part_name,
                        "description": network.name,
                        "netmaskLength": subnet_ip_mask[1],
                        "gateway": gw_ip}
        if dhcp_range:
            network_info["dhcpScope"] = dhcp_scopes

        if self._is_iplus:
            # Need to add the vrf name to the network info
            prof = self._config_profile_get(network.config_profile)
            if prof and prof.get('profileSubType') == 'network:universal':
                # For universal profile vrf has to e organization:partition
                network_info["vrfName"] = ':'.join((tenant_name,
                                                    self._part_name))
            else:
                # Otherwise, it should be left empty.
                network_info["vrfName"] = ""

        LOG.info("Creating %s network in DCNM.", network_info)

        res = self._create_network(network_info)
        if res and res.status_code in self._resp_ok:
            LOG.info("Created %s network in DCNM.", network_info)
        else:
            LOG.error("Failed to create %s network in DCNM.",
                      network_info)
            raise dexc.DfaClientRequestFailed(reason=res)

    def create_service_network(self, tenant_name, network, subnet,
                               dhcp_range=True):
        """Create network on the DCNM.

        :param tenant_name: name of tenant the network belongs to
        :param network: network parameters
        :param subnet: subnet parameters of the network
        """
        network_info = {}
        subnet_ip_mask = subnet.cidr.split('/')
        if self._default_md is None:
            self._set_default_mobility_domain()
        vlan_id = '0'
        gw_ip = subnet.gateway_ip
        part_name = network.part_name
        if not part_name:
            part_name = self._part_name

        if network.vlan_id:
            vlan_id = str(network.vlan_id)
            if network.mob_domain_name is not None:
                mob_domain_name = network.mob_domain_name
            else:
                mob_domain_name = self._default_md
        else:
            mob_domain_name = None

        seg_id = str(network.segmentation_id)
        seg_str = "$segmentId=" + seg_id
        cfg_args = [
            seg_str,
            "$netMaskLength=" + subnet_ip_mask[1],
            "$gatewayIpAddress=" + gw_ip,
            "$networkName=" + network.name,
            "$vlanId=" + vlan_id,
            "$vrfName=" + tenant_name + ':' + part_name
        ]
        cfg_args = ';'.join(cfg_args)

        ip_range = ','.join(["%s-%s" % (p['start'], p['end']) for p in
                             subnet.allocation_pools])

        dhcp_scopes = {'ipRange': ip_range,
                       'subnet': subnet.cidr,
                       'gateway': gw_ip}

        network_info = {"vlanId": vlan_id,
                        "mobilityDomainId": mob_domain_name,
                        "profileName": network.config_profile,
                        "networkName": network.name,
                        "configArg": cfg_args,
                        "organizationName": tenant_name,
                        "partitionName": part_name,
                        "description": network.name,
                        "netmaskLength": subnet_ip_mask[1],
                        "gateway": gw_ip}
        if seg_id:
            network_info["segmentId"] = seg_id
        if dhcp_range:
            network_info["dhcpScope"] = dhcp_scopes
        if hasattr(subnet, 'secondary_gw'):
            network_info["secondaryGateway"] = subnet.secondary_gw
        if self._is_iplus:
            # Need to add the vrf name to the network info
            prof = self._config_profile_get(network.config_profile)
            if prof and prof.get('profileSubType') == 'network:universal':
                # For universal profile vrf has to e organization:partition
                network_info["vrfName"] = ':'.join((tenant_name, part_name))
            else:
                # Otherwise, it should be left empty.
                network_info["vrfName"] = ""

        LOG.info("Creating %s network in DCNM.", network_info)

        res = self._create_network(network_info)
        if res and res.status_code in self._resp_ok:
            LOG.info("Created %s network in DCNM.", network_info)
        else:
            LOG.error("Failed to create %s network in DCNM.",
                      network_info)
            raise dexc.DfaClientRequestFailed(reason=self._failure_msg(res))

    def delete_network(self, tenant_name, network):
        """Delete network on the DCNM.

        :param tenant_name: name of tenant the network belongs to
        :param network: object that contains network parameters
        """
        seg_id = network.segmentation_id
        network_info = {
            'organizationName': tenant_name,
            'partitionName': self._part_name,
            'segmentId': seg_id,
        }
        LOG.debug("Deleting %s network in DCNM.", network_info)

        res = self._delete_network(network_info)
        if res and res.status_code in self._resp_ok:
            LOG.debug("Deleted %s network in DCNM.", network_info)
        else:
            LOG.error("Failed to delete %s network in DCNM.",
                      network_info)
            raise dexc.DfaClientRequestFailed(reason=res)

    def delete_service_network(self, tenant_name, network):
        """Delete service network on the DCNM.

        :param tenant_name: name of tenant the network belongs to
        :param network: object that contains network parameters
        """

        network_info = {}
        part_name = network.part_name
        if not part_name:
            part_name = self._part_name
        seg_id = str(network.segmentation_id)
        if network.vlan:
            vlan_id = str(network.vlan)
            if network.mob_domain_name is not None:
                mob_domain_name = network.mob_domain_name
            else:
                # The current way will not work since _default_md is obtained
                # during create_service_network. It's preferrable to get it
                # during init TODO(padkrish)
                if self._default_md is None:
                    self._set_default_mobility_domain()
                mob_domain_name = self._default_md
            network_info = {
                'organizationName': tenant_name,
                'partitionName': part_name,
                'mobDomainName': mob_domain_name,
                'vlanId': vlan_id,
                'segmentId': seg_id,
            }
        else:
            network_info = {
                'organizationName': tenant_name,
                'partitionName': part_name,
                'segmentId': seg_id,
            }
        LOG.debug("Deleting %s network in DCNM.", network_info)

        res = self._delete_network(network_info)
        if res and res.status_code in self._resp_ok:
            LOG.debug("Deleted %s network in DCNM.", network_info)
        else:
            LOG.error("Failed to delete %s network in DCNM.",
                      network_info)
            raise dexc.DfaClientRequestFailed(reason=self._failure_msg(res))

    def delete_project(self, tenant_name, part_name):
        """Delete project on the DCNM.

        :param tenant_name: name of project.
        :param part_name: name of partition.
        """
        res = self._delete_partition(tenant_name, part_name)
        if res and res.status_code in self._resp_ok:
            LOG.debug("Deleted %s partition in DCNM.", part_name)
        else:
            LOG.error("Failed to delete %(part)s partition in DCNM."
                      "Response: %(res)s", {'part': part_name, 'res': res})
            raise dexc.DfaClientRequestFailed(reason=res)

        res = self._delete_org(tenant_name)
        if res and res.status_code in self._resp_ok:
            LOG.debug("Deleted %s organization in DCNM.", tenant_name)
        else:
            LOG.error("Failed to delete %(org)s organization in DCNM."
                      "Response: %(res)s", {'org': tenant_name, 'res': res})
            raise dexc.DfaClientRequestFailed(reason=res)

    def delete_partition(self, org_name, partition_name):
        """Send partition delete request to DCNM.

        :param partition_name: name of partition to be deleted
        """
        res = self._delete_partition(org_name, partition_name)
        if res and res.status_code in self._resp_ok:
            LOG.debug("Deleted %s partition in DCNM.", partition_name)
        else:
            LOG.error("Failed to delete %(part)s partition in DCNM."
                      "Response: %(res)s",
                      ({'part': partition_name, 'res': res}))
            raise dexc.DfaClientRequestFailed(reason=self._failure_msg(res))

    def create_project(self, orch_id, org_name, part_name, dci_id, desc=None):
        """Create project on the DCNM.

        :param orch_id: orchestrator ID
        :param org_name: name of organization.
        :param part_name: name of partition.
        :param dci_id: Data Center interconnect id.
        :param desc: description of project.
        """
        desc = desc or org_name
        res = self._create_org(orch_id, org_name, desc)
        if res and res.status_code in self._resp_ok:
            LOG.debug("Created %s organization in DCNM.", org_name)
        else:
            LOG.error("Failed to create %(org)s organization in DCNM."
                      "Response: %(res)s", {'org': org_name, 'res': res})
            raise dexc.DfaClientRequestFailed(reason=res)

        self.create_partition(org_name, part_name, dci_id,
                              self.default_vrf_profile, desc=desc)

    def update_project(self, org_name, part_name, dci_id=UNKNOWN_DCI_ID,
                       service_node_ip=UNKNOWN_SRVN_NODE_IP,
                       vrf_prof=None, desc=None):
        """Update project on the DCNM.

        :param org_name: name of organization.
        :param part_name: name of partition.
        :param dci_id: Data Center interconnect id.
        :param desc: description of project.
        """
        desc = desc or org_name
        res = self._create_or_update_partition(org_name, part_name, desc,
                                               dci_id=dci_id,
                                               service_node_ip=service_node_ip,
                                               vrf_prof=vrf_prof,
                                               operation='PUT')
        if res and res.status_code in self._resp_ok:
            LOG.debug("Update %s partition in DCNM.", part_name)
        else:
            LOG.error("Failed to update %(part)s partition in DCNM."
                      "Response: %(res)s", {'part': part_name, 'res': res})
            raise dexc.DfaClientRequestFailed(reason=res)

    def create_partition(self, org_name, part_name, dci_id, vrf_prof,
                         service_node_ip=None, desc=None):
        """Create partition on the DCNM.

        :param org_name: name of organization to be created
        :param part_name: name of partition to be created
        :param dci_id: DCI ID
        :vrf_prof: VRF profile for the partition
        :param service_node_ip: Specifies the Default route IP address.
        :param desc: string that describes organization
        """
        desc = desc or org_name
        res = self._create_or_update_partition(org_name, part_name,
                                               desc, dci_id=dci_id,
                                               service_node_ip=service_node_ip,
                                               vrf_prof=vrf_prof)
        if res and res.status_code in self._resp_ok:
            LOG.debug("Created %s partition in DCNM.", part_name)
        else:
            LOG.error("Failed to create %(part)s partition in DCNM."
                      "Response: %(res)s", ({'part': part_name, 'res': res}))
            raise dexc.DfaClientRequestFailed(reason=self._failure_msg(res))

    def get_partition_vrfProf(self, org_name, part_name=None, part_info=None):
        """get VRF Profile for the partition from the DCNM.

        :param org_name: name of organization
        :param part_name: name of partition
        """
        vrf_profile = None
        if part_info is None:
            part_info = self._get_partition(org_name, part_name)
            LOG.info("query result from dcnm for partition info is %s",
                     part_info)
        if ("vrfProfileName" in part_info):
            vrf_profile = part_info.get("vrfProfileName")
        return vrf_profile

    def get_partition_dciId(self, org_name, part_name, part_info=None):
        """get DCI ID for the partition.

        :param org_name: name of organization
        :param part_name: name of partition
        """
        if part_info is None:
            part_info = self._get_partition(org_name, part_name)
            LOG.info("query result from dcnm for partition info is %s",
                     part_info)
        if part_info is not None and "dciId" in part_info:
            return part_info.get("dciId")

    def get_partition_serviceNodeIp(self, org_name, part_name, part_info=None):
        """get Service Node IP address from the DCNM.

        :param org_name: name of organization
        :param part_name: name of partition
        """
        if part_info is None:
            part_info = self._get_partition(org_name, part_name)
            LOG.info("query result from dcnm for partition info is %s",
                     part_info)
        if part_info is not None and "serviceNodeIpAddress" in part_info:
            return part_info.get("serviceNodeIpAddress")

    def get_partition_segmentId(self, org_name, part_name, part_info=None):
        """get partition Segment ID from the DCNM.

        :param org_name: name of organization
        :param part_name: name of partition
        """
        if part_info is None:
            part_info = self._get_partition(org_name, part_name)
            LOG.info("query result from dcnm for partition info is %s",
                     part_info)
        if part_info is not None and "partitionSegmentId" in part_info:
            return part_info.get("partitionSegmentId")

    def list_networks(self, org, part):
        """Return list of networks from DCNM.

        :param org: name of organization.
        :param part: name of partition.
        """
        if org and part:
            list_url = self._del_part + '/networks'
            list_url = list_url % (org, part)
            res = self._send_request('GET', list_url, '', 'networks')
            if res and res.status_code in self._resp_ok:
                return res.json()

    def list_organizations(self):
        """Return list of organizations from DCNM."""

        try:
            res = self._send_request('GET', self._org_url, '', 'organizations')
            if res and res.status_code in self._resp_ok:
                return res.json()
        except dexc.DfaClientRequestFailed:
            LOG.error("Failed to send request to DCNM.")

    def get_network(self, org, segid):
        """Return given network from DCNM.

        :param org: name of organization.
        :param segid: segmentation id of the network.
        """
        network_info = {
            'organizationName': org,
            'partitionName': self._part_name,
            'segmentId': segid,
        }
        res = self._get_network(network_info)
        if res and res.status_code in self._resp_ok:
            return res.json()

    def get_version(self):
        """Get the DCNM version."""

        url = '%s://%s/rest/dcnm-version' % (self.dcnm_protocol, self._ip)
        payload = {}

        try:
            res = self._send_request('GET', url, payload, 'dcnm-version')
            if res and res.status_code in self._resp_ok:
                return res.json().get('Dcnm-Version')
        except dexc.DfaClientRequestFailed as exc:
            LOG.error("Failed to get DCNM version.")
            sys.exit("ERROR: Failed to connect to DCNM: %s", exc)

    def _verify_protocol(self, protocol):
        try:
            self._login_request("%s://%s/rest/logon" % (protocol, self._ip))
            self._logout_request("%s://%s/rest/logout" % (protocol, self._ip))
        except (requests.HTTPError, requests.Timeout,
                requests.ConnectionError) as exc:
            LOG.error("Login Test failed for %(protocol)s Exc %(exc)s.",
                      {'protocol': protocol, 'exc': exc})
            return False
        return True

    def get_dcnm_protocol(self):
        """Routine to find out if DCNM is using http or https.

        DCNM 10 (Fuji-4) and above does not support http. Only https is
        supported and enabled by default.
        Prior DCNM versions supported both http and https. But, only http
        was enabled by default.
        So, enabler needs to find out if DCNM is supporting http or https to
        be friendly with the existing installed setups.
        """
        if self._verify_protocol('https'):
            return 'https'
        if self._verify_protocol('http'):
            return 'http'
        sys.exit("ERROR: Both http and https test failed")

    def _build_url(self, url_remaining):
        """This function builds the URL from host, protocol and string. """
        return self.host_protocol_url + url_remaining

    def fill_urls(self):
        """This assigns the URL's based on the protocol. """

        protocol = self.dcnm_protocol
        self._org_url = '%s://%s/rest/auto-config/organizations' % (
            (protocol, self._ip))
        self._create_network_url = ('%s://%s/' % (protocol, self._ip) +
                                    'rest/auto-config/organizations'
                                    '/%s/partitions/%s/networks')
        self.host_protocol_url = '%s://%s/' % (protocol, self._ip)
        self._create_network_url = self._build_url(
            'rest/auto-config/organizations'
            '/%s/partitions/%s/networks')
        self._cfg_profile_list_url = '%s://%s/rest/auto-config/profiles' % (
            (protocol, self._ip))
        self._cfg_profile_get_url = self._cfg_profile_list_url + '/%s'
        self._global_settings_url = self._build_url(
            'rest/auto-config/settings')
        self._create_part_url = self._build_url(
            'rest/auto-config/organizations/%s/partitions')
        self._update_part_url = self._build_url(
            'rest/auto-config/organizations/%s/partitions/%s')
        self._del_org_url = self._build_url(
            'rest/auto-config/organizations/%s')
        self._del_part = self._build_url(
            'rest/auto-config/organizations/%s/partitions/%s')
        self._network_url = self._build_url(
            'rest/auto-config/organizations/%s/partitions/'
            '%s/networks/segment/%s')
        self._network_mob_url = self._build_url(
            'rest/auto-config/organizations/%s/partitions/'
            '%s/networks/vlan/%s/mobility-domain/%s')
        self._segmentid_ranges_url = self._build_url(
            'rest/settings/segmentid-ranges')
        self._login_url = self._build_url('rest/logon')
        self._logout_url = self._build_url('rest/logout')
